from itertools import cycle

from streamparse import Spout


class WordSpout(Spout):
    outputs = ['word']

    def initialize(self, stormconf, context):
        self.words = cycle(['dog', 'cat', 'zebra', 'elephant'])

    def next_tuple(self):
        word = next(self.words)
        self.emit([word])



"""
Spout Specification

This module is called spout to mirror organization of storm package.
"""
from __future__ import absolute_import

from .component import JavaComponentSpec, ShellComponentSpec


class ShellSpoutSpec(ShellComponentSpec):
    def __init__(self, component_cls, name=None, command=None, script=None,
                 par=1, config=None, outputs=None):
        super(ShellSpoutSpec, self).__init__(component_cls,
                                             name=name, par=par,
                                             config=config, outputs=outputs,
                                             command=command, script=script)


class JavaSpoutSpec(JavaComponentSpec):
    def __init__(self, component_cls, name=None, serialized_java=None,
                 full_class_name=None, args_list=None, par=1, config=None,
                 outputs=None):
        super(JavaSpoutSpec, self).__init__(component_cls,
                                            name=name, par=par,
                                            config=config, outputs=outputs,
                                            serialized_java=serialized_java,
                                            full_class_name=full_class_name,
                                            args_list=args_list)



"""
Tail the specified log files.
"""

from __future__ import absolute_import, print_function

from fabric.api import env, execute, parallel, run

from .common import add_environment, add_name, add_pattern
from ..util import activate_env, get_logfiles_cmd, get_topology_definition


@parallel
def _tail_logs(topology_name, pattern, follow, num_lines):
    """
    Actual task to run tail on all servers in parallel.
    """
    ls_cmd = get_logfiles_cmd(topology_name=topology_name, pattern=pattern)
    tail_pipe = " | xargs tail -n {}".format(num_lines)
    if follow:
        tail_pipe += " -f"
    run(ls_cmd + tail_pipe)


def tail_topology(topology_name=None, env_name=None, pattern=None, follow=False,
                  num_lines=10):
    """Follow (tail -f) the log files on remote Storm workers.

    Will use the `log_path` and `workers` properties from config.json.
    """
    topology_name = get_topology_definition(topology_name)[0]
    activate_env(env_name)
    execute(_tail_logs, topology_name, pattern, follow, num_lines,
            hosts=env.storm_workers)


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser('tail',
                                      description=__doc__,
                                      help=main.__doc__)
    subparser.set_defaults(func=main)
    add_environment(subparser)
    subparser.add_argument('-f', '--follow',
                           action='store_true',
                           help='Keep files open and append output as they '
                                'grow.  This is the same as "-f" parameter for '
                                'the tail command that will be executed on the '
                                'Storm workers.')
    subparser.add_argument('-l', '--num_lines',
                           default=10,
                           help='tail outputs the last NUM_LINES lines of the '
                                'logs. (default: %(default)s)')
    add_name(subparser)
    add_pattern(subparser)


def main(args):
    """ Tail logs for specified Storm topology. """
    tail_topology(topology_name=args.name, env_name=args.environment,
                  pattern=args.pattern, follow=args.follow,
                  num_lines=args.num_lines)



"""
Word count topology
"""

from streamparse import Grouping, Topology

from bolts.wordcount import WordCountBolt
from spouts.words import WordSpout


class WordCount(Topology):
    word_spout = WordSpout.spec()
    count_bolt = WordCountBolt.spec(inputs={word_spout: Grouping.fields('word')},
                                    par=2)



"""
Run a local Storm topology.

Note: If you have "org.apache.storm" in your uberjar-exclusions list in your
project.clj, this will fail. Temporarily remove it to use `sparse run`.
"""

from __future__ import absolute_import, print_function

from argparse import RawDescriptionHelpFormatter
from tempfile import NamedTemporaryFile

from fabric.api import local
from ruamel import yaml

from ..util import (get_env_config, get_topology_definition,
                    get_topology_from_file, local_storm_version,
                    storm_lib_version)
from .common import (add_ackers, add_debug, add_environment, add_name,
                     add_options, add_workers, resolve_options)
from .jar import jar_for_deploy


def run_local_topology(name=None, env_name=None, time=0, options=None):
    """Run a topology locally using Flux and `storm jar`."""
    name, topology_file = get_topology_definition(name)
    env_name, env_config = get_env_config(env_name)
    topology_class = get_topology_from_file(topology_file)

    storm_options = resolve_options(options, env_config, topology_class, name)
    if storm_options['topology.acker.executors'] != 0:
        storm_options['topology.acker.executors'] = 1
    storm_options['topology.workers'] = 1

    # Check Storm version is the same
    local_version = local_storm_version()
    project_version = storm_lib_version()
    if local_version != project_version:
        raise ValueError('Local Storm version, {}, is not the same as the '
                         'version in your project.clj, {}. The versions must '
                         'match.'.format(local_version, project_version))

    # Prepare a JAR that has Storm dependencies packaged
    topology_jar = jar_for_deploy(simple_jar=False)

    if time <= 0:
        time = 9223372036854775807  # Max long value in Java

    # Write YAML file
    with NamedTemporaryFile(mode='w', suffix='.yaml', delete=False) as yaml_file:
        topology_flux_dict = topology_class.to_flux_dict(name)
        topology_flux_dict['config'] = storm_options
        yaml.safe_dump(topology_flux_dict, yaml_file)
        cmd = ('storm jar {jar} org.apache.storm.flux.Flux --local --no-splash '
               '--sleep {time} {yaml}'.format(jar=topology_jar,
                                              time=time,
                                              yaml=yaml_file.name))
        local(cmd)


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser('run',
                                      description=__doc__,
                                      help=main.__doc__,
                                      formatter_class=RawDescriptionHelpFormatter)
    subparser.set_defaults(func=main)
    add_ackers(subparser)
    add_debug(subparser)
    add_environment(subparser)
    add_name(subparser)
    add_options(subparser)
    subparser.add_argument('-t', '--time',
                           default=0,
                           type=int,
                           help='Time (in seconds) to keep local cluster '
                                'running. If time <= 0, run indefinitely. '
                                '(default: %(default)s)')
    add_workers(subparser)


def main(args):
    """Run the local topology with the given arguments"""
    run_local_topology(name=args.name, time=args.time, options=args.options,
                       env_name=args.environment)



"""
Display uptime for workers in running Storm topologies.
"""

from __future__ import absolute_import, print_function

from pkg_resources import parse_version
from prettytable import PrettyTable
from six import iteritems, itervalues

from .common import add_environment
from ..util import get_ui_json, get_ui_jsons, storm_lib_version


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser('worker_uptime',
                                      description=__doc__,
                                      help=main.__doc__)
    subparser.set_defaults(func=main)
    add_environment(subparser)


def display_worker_uptime(env_name):
    topology_summary_path = '/api/v1/topology/summary'
    topology_detail_path = '/api/v1/topology/{topology}'
    component_path = '/api/v1/topology/{topology}/component/{component}'
    topo_summary_json = get_ui_json(env_name, topology_summary_path)
    topology_ids = [x['id'] for x in topo_summary_json['topologies']]
    topology_components = dict()
    worker_stats = []
    topology_detail_jsons = get_ui_jsons(env_name,
                                         (topology_detail_path.format(topology=topology)
                                          for topology in topology_ids))

    for topology in topology_ids:
        topology_detail_json = topology_detail_jsons[topology_detail_path.format(topology=topology)]
        spouts = [x['spoutId'] for x in topology_detail_json['spouts']]
        bolts = [x['boltId'] for x in topology_detail_json['bolts']]
        topology_components[topology] = spouts + bolts

    comp_details = get_ui_jsons(env_name,
                                (component_path.format(topology=topology,
                                                       component=comp)
                                 for topology, comp_list in iteritems(topology_components)
                                 for comp in comp_list))

    for comp_detail in itervalues(comp_details):
        worker_stats += [(worker['host'], worker['id'], worker['uptime'],
                          worker['workerLogLink']) for worker in
                         comp_detail['executorStats']]
    worker_stats = sorted(set(worker_stats))

    print("# Worker Stats")
    table = PrettyTable(["Host", "Worker ID", "Uptime", "Log URL"])
    table.align = 'l'
    table.align['Uptime'] = 'r'
    for row in worker_stats:
        table.add_row(row)
    print(table)
    print()


def main(args):
    """ Display uptime for Storm workers. """
    storm_version = storm_lib_version()
    if storm_version >= parse_version('0.9.2-incubating'):
        display_worker_uptime(args.environment)
    else:
        print("ERROR: Storm {0} does not support this command."
              .format(storm_version))



"""
Create or update a virtualenv on the Storm workers.  This will be done
automatically upon submit, but this command is provided to help with testing
and debugging.
"""

from __future__ import absolute_import, print_function, unicode_literals

import os
from io import open

from fabric.api import env, execute, parallel, prefix, put, puts, run, show
from fabric.contrib.files import exists

from .common import add_environment, add_name, add_override_name
from ..util import (activate_env, die, get_config, get_env_config,
                    get_topology_definition)


@parallel
def _create_or_update_virtualenv(virtualenv_root,
                                 virtualenv_name,
                                 requirements_file,
                                 virtualenv_flags=None):
    with show('output'):
        virtualenv_path = '/'.join((virtualenv_root, virtualenv_name))
        if not exists(virtualenv_path):
            if virtualenv_flags is None:
                virtualenv_flags = ''
            puts("virtualenv not found in {}, creating one.".format(virtualenv_root))
            run("virtualenv {} {}".format(virtualenv_path, virtualenv_flags))

        puts("Uploading requirements.txt to temporary file.")
        tmpfile = run("mktemp /tmp/streamparse_requirements-XXXXXXXXX.txt")
        put(requirements_file, tmpfile)

        puts("Updating virtualenv: {}".format(virtualenv_name))
        cmd = "source {}".format(os.path.join(virtualenv_path, 'bin/activate'))
        with prefix(cmd):
            run("pip install -r {} --exists-action w".format(tmpfile))

        run("rm {}".format(tmpfile))


def create_or_update_virtualenvs(env_name, topology_name, override_name=None):
    """Create or update virtualenvs on remote servers.

    Assumes that virtualenv is on the path of the remote server(s).

    :param env_name: the name of the environment in config.json.
    :param topology_name: the name of the topology (and virtualenv).
    :param override_name: the name that we should use for the virtualenv, even
                          though the topology file has a different name.
    """
    config = get_config()
    topology_name = get_topology_definition(topology_name)[0]
    env_name, env_config = get_env_config(env_name)
    if override_name is None:
        override_name = topology_name

    config["virtualenv_specs"] = config["virtualenv_specs"].rstrip("/")

    requirements_path = os.path.join(config["virtualenv_specs"],
                                     '{}.txt'.format(topology_name))

    # Setup the fabric env dictionary
    activate_env(env_name)
    # Check to ensure streamparse is in requirements
    with open(requirements_path, "r") as fp:
        found_streamparse = False
        for line in fp:
            if "streamparse" in line:
                found_streamparse = True
                break

        if not found_streamparse:
            die("Could not find streamparse in your requirements file ({}). "
                "streamparse is required for all topologies."
                .format(requirements_path))

    execute(_create_or_update_virtualenv, env.virtualenv_root, override_name,
            requirements_path,
            virtualenv_flags=env_config.get('virtualenv_flags'),
            hosts=env.storm_workers)


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser('update_virtualenv',
                                      description=__doc__,
                                      help=main.__doc__)
    subparser.set_defaults(func=main)
    add_environment(subparser)
    add_name(subparser)
    add_override_name(subparser)


def main(args):
    """ Create or update a virtualenv on Storm workers. """
    create_or_update_virtualenvs(args.environment, args.name,
                                 override_name=args.override_name)



"""
Module to add streamparse-specific extensions to pystorm Component classes
"""
import pystorm
from pystorm.component import StormHandler  # This is used by other code


class Component(pystorm.component.Component):
    """pystorm Component with streamparse-specific additions

    :ivar outputs: The outputs
    :ivar config: Component-specific config settings to pass to Storm.
    """
    outputs = None
    par = 1
    config = None

    @classmethod
    def spec(cls, *args, **kwargs):
        """This method exists only to give a more informative error message."""
        raise TypeError('Specifications should either be bolts or spouts. '
                        'Given: {!r}'.format(cls))



"""
This module provides the base sparse command and a load hook for dynamically
adding other subcommands.  The "load_suparsers" function searches for modules
in the streamparse/bin directory that have a "subparser_hook" method. The
"subparser_hook" accepts a the sparse subparsers object and adds it's
subparser as needed.
"""

from __future__ import absolute_import

import argparse
import importlib
import os
import pkgutil
import sys

from ..version import __version__


def load_subparsers(subparsers):
    """
    searches modules in streamparse/bin for a 'subparser_hook' method and calls
    the 'subparser_hook' method on the sparse subparsers object.
    """
    for _, mod_name, is_pkg in pkgutil.iter_modules([os.path.dirname(__file__)]):
        if not is_pkg and mod_name not in sys.modules:
            module = importlib.import_module('streamparse.cli.{}'.format(mod_name))
            # check for the subparser hook
            if hasattr(module, 'subparser_hook'):
                module.subparser_hook(subparsers)


def main():
    """main entry point for sparse"""
    parser = argparse.ArgumentParser(description='Utilities for managing Storm'
                                                 '/streamparse topologies.',
                                     epilog='sparse provides a front-end to '
                                            'streamparse, a framework for '
                                            'creating Python projects for '
                                            'running, debugging, and '
                                            'submitting computation topologies '
                                            'against real-time streams, using '
                                            'Apache Storm. It requires java and'
                                            ' lein (Clojure build tool) to be '
                                            'on your $PATH, and uses lein and '
                                            'Clojure under the hood for JVM/'
                                            'Thrift interop.')
    subparsers = parser.add_subparsers(title='sub-commands')
    parser.add_argument('--version', action='version',
                        version='%(prog)s {}'.format(__version__))
    load_subparsers(subparsers)
    args = parser.parse_args()

    # http://grokbase.com/t/python/python-bugs-list/12arsq9ayf/issue16308-undocumented-behaviour-change-in-argparse-from-3-2-3-to-3-3-0
    if hasattr(args, 'func'):
        args.func(args)
    # python3.3+ argparse changes
    else:
        parser.print_help()
        sys.exit(1)

if __name__ == '__main__':
    main()



"""
Display stats about running Storm topologies.
"""

from __future__ import absolute_import, print_function

import sys
from itertools import chain

from pkg_resources import parse_version
from six import iteritems, string_types
from six.moves import map, zip

from ..util import (get_env_config, get_ui_json, get_ui_jsons,
                    print_stats_table, storm_lib_version)
from .common import add_environment, add_name



def display_stats(env_name, topology_name=None, component_name=None,
                  all_components=None):
    env_name = env_name
    if topology_name and all_components:
        _print_all_components(env_name, topology_name)
    elif topology_name and component_name:
        _print_component_status(env_name, topology_name, component_name)
    elif topology_name:
        _print_topology_status(env_name, topology_name)
    else:
        _print_cluster_status(env_name)


def _print_cluster_status(env_name):
    jsons = get_ui_jsons(env_name, ["/api/v1/cluster/summary",
                                    "/api/v1/topology/summary",
                                    "/api/v1/supervisor/summary"])
    # Print Cluster Summary
    ui_cluster_summary = jsons["/api/v1/cluster/summary"]
    columns = ['stormVersion', 'nimbusUptime', 'supervisors', 'slotsTotal',
               'slotsUsed', 'slotsFree', 'executorsTotal', 'tasksTotal']
    print_stats_table("Cluster summary", ui_cluster_summary, columns, 'r')
    # Print Topologies Summary
    ui_topologies_summary = jsons["/api/v1/topology/summary"]
    columns = ['name', 'id', 'status', 'uptime', 'workersTotal',
               'executorsTotal', 'tasksTotal']
    print_stats_table("Topology summary", ui_topologies_summary['topologies'],
                      columns, 'r')
    # Print Supervisor Summary
    ui_supervisor_summary = jsons["/api/v1/supervisor/summary"]
    columns = ['id', 'host', 'uptime', 'slotsTotal', 'slotsUsed']
    print_stats_table("Supervisor summary",
                      ui_supervisor_summary['supervisors'], columns, 'r',
                      {'host': 'l', 'uptime': 'l'})


def _get_topology_ui_detail(env_name, topology_name):
    env_name = get_env_config(env_name)[0]
    topology_id = _get_topology_id(env_name, topology_name)
    detail_url = '/api/v1/topology/%s' % topology_id
    detail = get_ui_json(env_name, detail_url)
    return detail


def _print_topology_status(env_name, topology_name):
    ui_detail = _get_topology_ui_detail(env_name, topology_name)
    # Print topology summary
    columns = ['name', 'id', 'status', 'uptime', 'workersTotal',
               'executorsTotal', 'tasksTotal']
    print_stats_table("Topology summary", ui_detail, columns, 'r')
    # Print topology stats
    columns = ['windowPretty', 'emitted', 'transferred', 'completeLatency',
               'acked', 'failed']
    print_stats_table("Topology stats", ui_detail['topologyStats'], columns,
                      'r')
    # Print spouts
    if ui_detail.get('spouts'):
        columns = ['spoutId', 'emitted', 'transferred', 'completeLatency',
                   'acked', 'failed']
        print_stats_table("Spouts (All time)", ui_detail['spouts'], columns,
                          'r', {'spoutId': 'l'})

    columns = ['boltId', 'executors', 'tasks', 'emitted', 'transferred',
               'capacity', 'executeLatency', 'executed', 'processLatency',
               'acked', 'failed', 'lastError']
    print_stats_table("Bolt (All time)", ui_detail['bolts'], columns, 'r',
                      {'boltId': 'l'})


def _get_component_ui_detail(env_name, topology_name, component_names):
    if isinstance(component_names, string_types):
        component_names = [component_names]
    env_name = get_env_config(env_name)[0]
    topology_id = _get_topology_id(env_name, topology_name)
    base_url = '/api/v1/topology/%s/component/%s'
    detail_urls = [base_url % (topology_id, name) for name in component_names]
    detail = get_ui_jsons(env_name, detail_urls)
    if len(detail) == 1:
        return detail.values()[0]
    else:
        return detail


def _print_all_components(env_name, topology_name):
    topology_ui_detail = _get_topology_ui_detail(env_name, topology_name)
    spouts = map(lambda spout: (spout['spoutId'],
                                topology_ui_detail.get('spouts', {})))
    bolts = map(lambda spout: (spout['boltId'],
                               topology_ui_detail.get('bolts', {})))
    ui_details = _get_component_ui_detail(env_name, topology_name, chain(spouts,
                                                                         bolts))
    names_and_keys = zip(map(lambda ui_detail: (ui_detail['name'],
                                                ui_details.values())),
                         ui_details.keys())
    for component_name, key in names_and_keys:
        _print_component_status(env_name, topology_name,
                                component_name, ui_details[key])


def _print_component_status(env_name, topology_name, component_name,
                            ui_detail=None):
    if not ui_detail:
        ui_detail = _get_component_ui_detail(env_name, topology_name,
                                             component_name)
    _print_component_summary(ui_detail)
    if ui_detail.get("componentType") == "spout":
        _print_spout_stats(ui_detail)
        _print_spout_output_stats(ui_detail)
        _print_spout_executors(ui_detail)
    elif ui_detail.get("componentType") == "bolt":
        _print_bolt_stats(ui_detail)
        _print_input_stats(ui_detail)
        _print_bolt_output_stats(ui_detail)


def _print_component_summary(ui_detail):
    columns = ['id', 'name', 'executors', 'tasks']
    print_stats_table("Component summary", ui_detail, columns, 'r')


def _print_bolt_stats(ui_detail):
    columns = ['windowPretty', 'emitted', 'transferred', 'executeLatency',
               'executed', 'processLatency', 'acked', 'failed']

    print_stats_table("Bolt stats", ui_detail['boltStats'], columns, 'r',
                      {'windowPretty': 'l'})


def _print_input_stats(ui_detail):
    columns = ['component', 'stream', 'executeLatency', 'processLatency',
               'executed', 'acked', 'failed']
    if ui_detail['inputStats']:
        print_stats_table("Input stats (All time)", ui_detail['inputStats'],
                          columns, 'r', {'component': 'l'})


def _print_bolt_output_stats(ui_detail):
    if ui_detail['outputStats']:
        columns = ['stream', 'emitted', 'transferred']
        print_stats_table("Output stats (All time)", ui_detail['outputStats'],
                          columns, 'r', {'stream': 'l'})


def _print_spout_stats(ui_detail):
    columns = ['windowPretty', 'emitted', 'transferred', 'completeLatency',
               'acked', 'failed']
    data = ui_detail['spoutSummary'][-1].copy()
    print_stats_table("Spout stats", data, columns, 'r', {'windowPretty': 'l'})


def _print_spout_output_stats(ui_detail):
    columns = ['stream', 'emitted', 'transferred', 'completeLatency',
               'acked', 'failed']
    print_stats_table("Output stats (All time)", ui_detail['outputStats'],
                      columns, 'r', {'stream': 'l'})


def _print_spout_executors(ui_detail):
    columns = ['id', 'uptime', 'host', 'port', 'emitted',
               'transferred', 'completeLatency', 'acked', 'failed']
    print_stats_table("Executors (All time)", ui_detail['executorStats'],
                      columns, 'r', {'host': 'l'})


def _get_topology_id(env_name, topology_name):
    """Get toplogy ID from summary json provided by UI api
    """
    summary_url = '/api/v1/topology/summary'
    topology_summary = get_ui_json(env_name, summary_url)
    for topology in topology_summary["topologies"]:
        if topology_name == topology["name"]:
            return topology["id"]


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser('stats',
                                      description=__doc__,
                                      help=main.__doc__)
    subparser.set_defaults(func=main)
    subparser.add_argument('--all',
                           action='store_true',
                           help='All available stats.')
    subparser.add_argument('-c', '--component',
                           help='Topology component (bolt/spout) name as '
                                'specified in Clojure topology specification')
    add_environment(subparser)
    add_name(subparser)


def main(args):
    """ Display stats about running Storm topologies. """
    storm_version = storm_lib_version()
    if storm_version >= parse_version('0.9.2-incubating'):
        display_stats(args.environment, topology_name=args.name,
                      component_name=args.component, all_components=args.all)
    else:
        print("ERROR: Storm {0} does not support this command."
              .format(storm_version))
        sys.stdout.flush()



"""
Remove all logs from Storm workers for the specified Storm topology.
"""

from __future__ import absolute_import, print_function

from fabric.api import env, execute, parallel, sudo

from .common import add_environment, add_name, add_pattern
from ..util import activate_env, get_topology_definition, get_logfiles_cmd


@parallel
def _remove_logs(topology_name, pattern, remove_worker_logs):
    """
    Actual task to remove logs on all servers in parallel.
    """
    ls_cmd = get_logfiles_cmd(topology_name=topology_name, pattern=pattern,
                              include_worker_logs=remove_worker_logs)
    rm_pipe = " | xargs rm"
    sudo(ls_cmd + rm_pipe, warn_only=True)


def remove_logs(topology_name=None, env_name=None, pattern=None,
                remove_worker_logs=False):
    """Remove all Python logs on Storm workers in the log.path directory."""
    topology_name = get_topology_definition(topology_name)[0]
    activate_env(env_name)
    execute(_remove_logs, topology_name, pattern, remove_worker_logs,
            hosts=env.storm_workers)


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser('remove_logs',
                                      description=__doc__,
                                      help=main.__doc__)
    subparser.set_defaults(func=main)
    add_environment(subparser)
    add_name(subparser)
    add_pattern(subparser)
    subparser.add_argument('-u', '--user',
                           help="User argument to sudo when deleting logs.")
    subparser.add_argument('-w', '--remove_worker_logs',
                           help='Remove not only topology-specific logs, but '
                                'also worker logs that may be shared between '
                                'topologies.',
                           action='store_true')

def main(args):
    """ Remove logs from Storm workers. """
    remove_logs(topology_name=args.name, env_name=args.environment,
                pattern=args.pattern,
                remove_worker_logs=args.remove_worker_logs)



"""
Bolt Specification

This module is called bolt to mirror organization of storm package.
"""
from __future__ import absolute_import

from .component import JavaComponentSpec, ShellComponentSpec


class ShellBoltSpec(ShellComponentSpec):
    def __init__(self, component_cls, name=None, command=None, script=None,
                 inputs=None, par=1, config=None, outputs=None):
        super(ShellBoltSpec, self).__init__(component_cls,
                                            name=name,
                                            inputs=inputs,
                                            par=par,
                                            config=config,
                                            outputs=outputs,
                                            command=command,
                                            script=script)


class JavaBoltSpec(JavaComponentSpec):
    def __init__(self, component_cls, name=None, serialized_java=None,
                 full_class_name=None, args_list=None, inputs=None,
                 par=1, config=None, outputs=None):
            super(JavaBoltSpec, self).__init__(component_cls,
                                               name=name,
                                               serialized_java=serialized_java,
                                               inputs=inputs,
                                               par=par,
                                               config=config, outputs=outputs,
                                               full_class_name=full_class_name,
                                               args_list=args_list)



"""
Tests for Topology DSL
"""

import json
import logging
import unittest
from io import BytesIO

from streamparse.dsl import Grouping, Stream, Topology
from streamparse.storm import (Bolt, Component, JavaBolt, JavaSpout, ShellBolt,
                               ShellSpout, Spout)
from streamparse.thrift import storm_thrift
from storm_thrift import JavaObjectArg

log = logging.getLogger(__name__)


class WordSpout(Spout):
    outputs = ["word"]


class WordCountBolt(Bolt):
    outputs = ["word", "count"]


class MultiStreamWordCountBolt(Bolt):
    outputs = [Stream(fields=['word', 'count']),
               Stream(fields=['all_word_count'], name='sum')]


class DatabaseDumperBolt(Bolt):
    outputs = []


class TopologyTests(unittest.TestCase):
    def test_basic_spec(self):
        class WordCount(Topology):
            word_spout = WordSpout.spec(par=2)
            word_bolt = WordCountBolt.spec(inputs={word_spout:
                                                   Grouping.fields("word")},
                                           par=8)

        self.assertEqual(len(WordCount.specs), 2)
        self.assertEqual(list(WordCount.word_bolt.inputs.keys()),
                         [WordCount.word_spout['default']])
        self.assertEqual(WordCount.thrift_spouts['word_spout'].common.parallelism_hint,
                         2)
        self.assertEqual(WordCount.thrift_bolts['word_bolt'].common.parallelism_hint,
                         8)
        self.assertEqual(WordCount.word_bolt.inputs[WordCount.word_spout['default']],
                         Grouping.fields('word'))

    def test_spec_with_inputs_as_list(self):
        class WordCount(Topology):
            word_spout = WordSpout.spec(par=2)
            word_bolt = WordCountBolt.spec(inputs=[word_spout], par=8)
        self.assertEqual(len(WordCount.specs), 2)
        self.assertEqual(len(WordCount.thrift_spouts), 1)
        self.assertEqual(len(WordCount.thrift_bolts), 1)
        self.assertEqual(list(WordCount.word_bolt.inputs.keys()),
                         [WordCount.word_spout['default']])
        self.assertEqual(WordCount.word_bolt.inputs[WordCount.word_spout['default']],
                         Grouping.SHUFFLE)

    def test_multi_stream_bolt(self):
        class WordCount(Topology):
            word_spout = WordSpout.spec(par=2)
            word_bolt = MultiStreamWordCountBolt.spec(inputs=[word_spout],
                                                      par=8)
            db_dumper_bolt = DatabaseDumperBolt.spec(par=4,
                                                     inputs=[word_bolt['sum'],
                                                             word_bolt['default']])
        self.assertEqual(len(WordCount.specs), 3)
        self.assertEqual(len(WordCount.thrift_spouts), 1)
        self.assertEqual(len(WordCount.thrift_bolts), 2)
        self.assertEqual(list(WordCount.word_bolt.inputs.keys()),
                         [WordCount.word_spout['default']])
        self.assertEqual(WordCount.word_bolt.inputs[WordCount.word_spout['default']],
                         Grouping.SHUFFLE)
        db_dumper_bolt_input_set = set(WordCount.db_dumper_bolt.inputs.keys())
        self.assertEqual(len(WordCount.db_dumper_bolt.inputs.keys()),
                         len(db_dumper_bolt_input_set))
        self.assertEqual(db_dumper_bolt_input_set,
                         {WordCount.word_bolt['sum'], WordCount.word_bolt['default']})

    def test_long_chain_spec(self):
        class WordCount(Topology):
            word_spout = WordSpout.spec()
            word_bolt1 = WordCountBolt.spec(inputs=[word_spout])
            word_bolt2 = WordCountBolt.spec(inputs=[word_bolt1])
            word_bolt3 = WordCountBolt.spec(inputs=[word_bolt2])
            word_bolt4 = WordCountBolt.spec(inputs=[word_bolt3])
            word_bolt5 = WordCountBolt.spec(inputs=[word_bolt4])
            word_bolt6 = WordCountBolt.spec(inputs=[word_bolt5])
            word_bolt7 = WordCountBolt.spec(inputs=[word_bolt6])
            word_bolt8 = WordCountBolt.spec(inputs=[word_bolt7])
            word_bolt9 = WordCountBolt.spec(inputs=[word_bolt8])
            word_bolt10 = WordCountBolt.spec(inputs=[word_bolt9])
            word_bolt11 = WordCountBolt.spec(inputs=[word_bolt10])
            word_bolt12 = WordCountBolt.spec(inputs=[word_bolt11])
        self.assertEqual(len(WordCount.specs), 13)
        self.assertEqual(len(WordCount.thrift_spouts), 1)
        self.assertEqual(len(WordCount.thrift_bolts), 12)
        self.assertEqual(list(WordCount.word_bolt1.inputs.keys())[0],
                         WordCount.word_spout['default'])
        self.assertEqual(WordCount.word_bolt1.inputs[WordCount.word_spout['default']],
                         Grouping.SHUFFLE)
        for i in range(2, 13):
            bolt = getattr(WordCount, 'word_bolt{}'.format(i))
            prev_bolt = getattr(WordCount, 'word_bolt{}'.format(i - 1))
            self.assertEqual(list(bolt.inputs.keys())[0], prev_bolt['default'])
            self.assertEqual(bolt.inputs[prev_bolt['default']], Grouping.SHUFFLE)

    def test_many_spouts_spec(self):
        class WordCount(Topology):
            word_spout1 = WordSpout.spec()
            word_spout2 = WordSpout.spec()
            word_spout3 = WordSpout.spec()
            word_spout4 = WordSpout.spec()
            word_spout5 = WordSpout.spec()
            word_bolt = WordCountBolt.spec(inputs=[word_spout1, word_spout2,
                                                   word_spout3, word_spout4,
                                                   word_spout5])

    def test_invalid_bolt_group_field(self):
        # Fields grouping must specify field output by spout
        with self.assertRaises(ValueError):
            class WordCount(Topology):
                word_spout = WordSpout.spec()
                word_bolt = WordCountBolt.spec(inputs={word_spout:
                                                       Grouping.fields("foo")})

    def test_empty_bolt_group_field(self):
        # Fields groupings require field names to be specified
        with self.assertRaises(ValueError):
            class WordCount(Topology):
                word_spout = WordSpout.spec()
                word_bolt = WordCountBolt.spec(inputs={word_spout:
                                                       Grouping.fields()})

    def test_duplicate_name(self):
        # Each component name must be unique
        with self.assertRaises(ValueError):
            class WordCount(Topology):
                word = WordSpout.spec()
                word_ = WordCountBolt.spec(name='word', inputs=[word])

    def test_no_input_bolt(self):
        # All bolts must have at least one input
        with self.assertRaises(ValueError):
            class WordCount(Topology):
                word_spout = WordSpout.spec()
                word_bolt = WordCountBolt.spec(inputs=[])

    def test_no_outputs_spout_empty(self):
        # All spouts must have output fields
        class PointlessSpout(Spout):
            outputs = []
        with self.assertRaises(ValueError):
            class WordCount(Topology):
                word_spout = PointlessSpout.spec()

    def test_no_outputs_spout(self):
        # All spouts must have output fields
        class PointlessSpout(Spout):
            outputs = []
        with self.assertRaises(ValueError):
            class WordCount(Topology):
                word_spout = PointlessSpout.spec()

    def test_base_component_rejection(self):
        # Topology components must inherit from Bolt/Spout, not Component directly
        class MyComponent(Component):
            outputs = []
        with self.assertRaises(TypeError):
            class WordCount(Topology):
                word_spout = MyComponent.spec()

    def test_component_instead_of_spec(self):
        # Make sure we catch things early when people forget to call .spec
        with self.assertRaises(TypeError):
            class WordCount(Topology):
                word_spout = WordSpout(input_stream=BytesIO(),
                                       output_stream=BytesIO())

    def test_no_spout(self):
        # Every topology must have a spout
        with self.assertRaises(ValueError):
            class WordCount(Topology):
                pass

    def test_zero_par(self):
        # Component parallelism (number of processes) can't be 0
        with self.assertRaises(ValueError):
            class WordCount(Topology):
                word_spout = WordSpout.spec(par=0)

    def test_negative_par(self):
        # Component parallelism (number of processes) can't be negative
        with self.assertRaises(ValueError):
            class WordCount(Topology):
                word_spout = WordSpout.spec(par=-1)

    def test_float_par(self):
        # Component parallelism (number of processes) must be an int
        with self.assertRaises(TypeError):
            class WordCount(Topology):
                word_spout = WordSpout.spec(par=5.4)

    def test_invalid_bolt_input_dict_key(self):
        # Keys in input dict must be either GlobalStreamId or ComponentSpec
        with self.assertRaises(TypeError):
            class WordCount(Topology):
                word_spout = WordSpout.spec()
                word_bolt = WordCountBolt.spec(inputs={'word_spout':
                                                       Grouping.fields('word')})

    def test_invalid_bolt_input_dict_val(self):
        # Values in input dict must be Grouping objects
        with self.assertRaises(TypeError):
            class WordCount(Topology):
                word_spout = WordSpout.spec()
                word_bolt = WordCountBolt.spec(inputs={word_spout:
                                                       'word'})

    def test_invalid_bolt_input_str(self):
        # inputs must be either list, dict, or None; not str
        with self.assertRaises(TypeError):
            class WordCount(Topology):
                word_spout = WordSpout.spec()
                word_bolt = WordCountBolt.spec(inputs='word_spout')

    def test_invalid_config_str(self):
        # configs must be either dict, or None; not str
        with self.assertRaises(TypeError):
            class WordCount(Topology):
                word_spout = WordSpout.spec()
                word_bolt = WordCountBolt.spec(inputs=[word_spout],
                                               config='{"foo": "bar"}')

    def test_invalid_topology_config_str(self):
        # configs must be either dict, or None; not str
        with self.assertRaises(TypeError):
            class WordCount(Topology):
                config = '{"foo": "bar"}'
                word_spout = WordSpout.spec()
                word_bolt = WordCountBolt.spec(inputs=[word_spout])

    def test_invalid_outputs_entry_int(self):
        # Outputs must either be list of strings or list of streams, not ints
        class BadSpout(Spout):
            outputs = ['foo', 2]
        with self.assertRaises(TypeError):
            class WordCount(Topology):
                bad_spout = BadSpout.spec()
                word_bolt = WordCountBolt.spec(inputs=[bad_spout])

    def test_invalid_outputs_str(self):
        # Outputs must either be list of strings or list of streams
        class BadSpout(Spout):
            outputs = 'foo'
        with self.assertRaises(TypeError):
            class WordCount(Topology):
                bad_spout = BadSpout.spec()
                word_bolt = WordCountBolt.spec(inputs=[bad_spout])

    def test_unknown_stream(self):
        # Should raise keyerror if stream name is not valid
        with self.assertRaises(KeyError):
            class WordCount(Topology):
                word_spout = WordSpout.spec()
                word_bolt = WordCountBolt.spec(inputs=[word_spout['word']])

    def test_perl_bolt(self):
        # Make sure ShellBolt.spec works
        class PerlWordCount(Topology):
            word_spout = WordSpout.spec(par=2)
            word_bolt = ShellBolt.spec(command='perl', script='count_words.pl',
                                       inputs={word_spout: Grouping.fields("word")},
                                       par=8, outputs=['word', 'count'])

        self.assertEqual(len(PerlWordCount.specs), 2)
        self.assertEqual(list(PerlWordCount.word_bolt.inputs.keys()),
                         [PerlWordCount.word_spout['default']])
        self.assertEqual(PerlWordCount.thrift_spouts['word_spout'].common.parallelism_hint,
                         2)
        self.assertEqual(PerlWordCount.thrift_bolts['word_bolt'].common.parallelism_hint,
                         8)
        self.assertEqual(PerlWordCount.word_bolt.inputs[PerlWordCount.word_spout['default']],
                         Grouping.fields('word'))

    def test_shell_bolt_no_command(self):
        # Should raise ValueError if command is None
        with self.assertRaises(ValueError):
            class PerlWordCount(Topology):
                word_spout = WordSpout.spec(par=2)
                word_bolt = ShellBolt.spec(script='count_words.pl',
                                           inputs={word_spout: Grouping.fields("word")},
                                           par=8, outputs=['word', 'count'])

    def test_shell_bolt_no_script(self):
        # Should raise TypeError if script is None
        with self.assertRaises(TypeError):
            class PerlWordCount(Topology):
                word_spout = WordSpout.spec(par=2)
                word_bolt = ShellBolt.spec(command='perl',
                                           inputs={word_spout: Grouping.fields("word")},
                                           par=8, outputs=['word', 'count'])

    def test_java_bolt_valid_arg_list(self):
        # JavaBolt should work fine when given basic data types
        class JavaWordCount(Topology):
            word_spout = WordSpout.spec(par=2)
            word_bolt = JavaBolt.spec(full_class_name='com.bar.foo.counter.WordCountBolt',
                                      args_list=[u'foo', 1, b'\x09\x10', True,
                                                 3.14159],
                                      inputs={word_spout: Grouping.fields("word")},
                                      par=8, outputs=['word', 'count'])
        java_object = JavaWordCount.thrift_bolts['word_bolt'].bolt_object.java_object
        self.assertEqual(java_object.full_class_name, 'com.bar.foo.counter.WordCountBolt')
        self.assertEqual(java_object.args_list,
                         [JavaObjectArg(string_arg=u'foo'),
                          JavaObjectArg(long_arg=1),
                          JavaObjectArg(binary_arg=b'\x09\x10'),
                          JavaObjectArg(bool_arg=True),
                          JavaObjectArg(double_arg=3.14159)])
        self.assertIsNone(JavaWordCount.thrift_bolts['word_bolt'].bolt_object.serialized_java)
        self.assertIsNone(JavaWordCount.thrift_bolts['word_bolt'].bolt_object.shell)

    def test_java_bolt_invalid_arg_list(self):
        # JavaBolt should raise TypeError when given something other than basic data types
        with self.assertRaises(TypeError):
            class JavaWordCount(Topology):
                word_spout = WordSpout.spec(par=2)
                word_bolt = JavaBolt.spec(full_class_name='com.bar.foo.counter.WordCountBolt',
                                          args_list=[{'foo': 'bar'}, 1],
                                          inputs={word_spout: Grouping.fields("word")},
                                          par=8, outputs=['word', 'count'])

    def test_java_bolt_valid_serialized_java(self):
        # JavaBolt should work fine when given a byte string for serialized_java
        serialized_java = b'\x01\x02\x03\x04\x05'
        class JavaWordCount(Topology):
            word_spout = WordSpout.spec(par=2)
            word_bolt = JavaBolt.spec(serialized_java=serialized_java,
                                      inputs={word_spout: Grouping.fields("word")},
                                      par=8, outputs=['word', 'count'])
        self.assertEqual(JavaWordCount.thrift_bolts['word_bolt'].bolt_object.serialized_java, serialized_java)
        self.assertIsNone(JavaWordCount.thrift_bolts['word_bolt'].bolt_object.java_object)
        self.assertIsNone(JavaWordCount.thrift_bolts['word_bolt'].bolt_object.shell)

    def test_perl_spout(self):
        # Make sure ShellBolt.spec works
        class PerlWordCount(Topology):
            word_spout = ShellSpout.spec(command='perl',
                                         script='count_words.pl', par=2,
                                         outputs=['word'])
            word_bolt = WordCountBolt.spec(inputs={word_spout: Grouping.fields("word")},
                                           par=8)
        self.assertEqual(len(PerlWordCount.specs), 2)
        self.assertEqual(list(PerlWordCount.word_bolt.inputs.keys()),
                         [PerlWordCount.word_spout['default']])
        self.assertEqual(PerlWordCount.thrift_spouts['word_spout'].common.parallelism_hint,
                         2)
        self.assertEqual(PerlWordCount.thrift_bolts['word_bolt'].common.parallelism_hint,
                         8)
        self.assertEqual(PerlWordCount.word_bolt.inputs[PerlWordCount.word_spout['default']],
                         Grouping.fields('word'))

    def test_shell_spout_no_command(self):
        # Should raise ValueError if command is None
        with self.assertRaises(ValueError):
            class PerlWordCount(Topology):
                word_spout = ShellSpout.spec(script='count_words.pl', par=8,
                                             outputs=['word'])
                word_bolt = WordCountBolt.spec(inputs={word_spout: Grouping.fields("word")},
                                               par=8)

    def test_shell_spout_no_script(self):
        # Should raise TypeError if script is None
        with self.assertRaises(TypeError):
            class PerlWordCount(Topology):
                word_spout = ShellSpout.spec(command='perl', par=8,
                                             outputs=['word'])
                word_bolt = WordCountBolt.spec(inputs={word_spout: Grouping.fields("word")},
                                               par=8)

    def test_java_spout_valid_arg_list(self):
        # JavaSpout should work fine when given basic data types
        class JavaWordCount(Topology):
            word_spout = JavaSpout.spec(full_class_name='com.bar.foo.counter.WordSpout',
                                        args_list=[u'foo', 1, b'\x09\x10', True,
                                                   3.14159],
                                        par=8, outputs=['word'])
            word_bolt = WordCountBolt.spec(inputs={word_spout: Grouping.fields("word")},
                                           par=8)
        java_object = JavaWordCount.thrift_spouts['word_spout'].spout_object.java_object
        self.assertEqual(java_object.full_class_name, 'com.bar.foo.counter.WordSpout')
        self.assertEqual(java_object.args_list,
                         [JavaObjectArg(string_arg=u'foo'),
                          JavaObjectArg(long_arg=1),
                          JavaObjectArg(binary_arg=b'\x09\x10'),
                          JavaObjectArg(bool_arg=True),
                          JavaObjectArg(double_arg=3.14159)])
        self.assertIsNone(JavaWordCount.thrift_spouts['word_spout'].spout_object.serialized_java)
        self.assertIsNone(JavaWordCount.thrift_spouts['word_spout'].spout_object.shell)

    def test_java_spout_invalid_arg_list(self):
        # JavaSpout should raise TypeError when given something other than basic data types
        with self.assertRaises(TypeError):
            class JavaWordCount(Topology):
                word_spout = JavaSpout.spec(full_class_name='com.bar.foo.counter.WordSpout',
                                            args_list=[{'foo': 'bar'}, 1],
                                            par=8, outputs=['word'])
                word_bolt = WordCountBolt.spec(inputs={word_spout: Grouping.fields("word")},
                                               par=8)

    def test_java_spout_valid_serialized_java(self):
        # JavaSpout should work fine when given a byte string for serialized_java
        serialized_java = b'\x01\x02\x03\x04\x05'
        class JavaWordCount(Topology):
            word_spout = JavaSpout.spec(serialized_java=serialized_java,
                                        par=8, outputs=['word'])
            word_bolt = WordCountBolt.spec(inputs={word_spout: Grouping.fields("word")},
                                           par=8)
        self.assertEqual(JavaWordCount.thrift_spouts['word_spout'].spout_object.serialized_java, serialized_java)
        self.assertIsNone(JavaWordCount.thrift_spouts['word_spout'].spout_object.java_object)
        self.assertIsNone(JavaWordCount.thrift_spouts['word_spout'].spout_object.shell)



"""
Utility functions for the DSL
"""

from __future__ import absolute_import

from six import integer_types, text_type

from ..thrift import storm_thrift


def to_java_arg(arg):
    """Converts Python objects to equivalent Thrift JavaObjectArgs"""
    if isinstance(arg, bool):
        java_arg = storm_thrift.JavaObjectArg(bool_arg=arg)
    elif isinstance(arg, integer_types):
        # Just use long all the time since Python 3 doesn't
        # distinguish between long and int
        java_arg = storm_thrift.JavaObjectArg(long_arg=arg)
    elif isinstance(arg, bytes):
        java_arg = storm_thrift.JavaObjectArg(binary_arg=arg)
    elif isinstance(arg, text_type):
        java_arg = storm_thrift.JavaObjectArg(string_arg=arg)
    elif isinstance(arg, float):
        java_arg = storm_thrift.JavaObjectArg(double_arg=arg)
    else:
        raise TypeError('Only basic data types can be specified'
                        ' as arguments to JavaObject '
                        'constructors.  Given: {!r}'
                        .format(arg))
    return java_arg


def to_python_arg(java_arg):
    """Convert a Thrift JavaObjectArg into a basic Python data type"""
    arg = None
    for val in java_arg.__dict__.values():
        if val is not None:
            arg = val
            break
    return arg


def to_python_arg_list(java_arg_list):
    """Convert a list of Thrift JavaObjectArg objects into a list of basic types"""
    return [to_python_arg(java_arg) for java_arg in java_arg_list]



"""
List the currently running Storm topologies.
"""

from __future__ import absolute_import

from ..util import (get_env_config, get_nimbus_client, print_stats_table,
                    ssh_tunnel)
from ..thrift import storm_thrift
from .common import add_environment


def _list_topologies(nimbus_client):
    """:returns: A list of running Storm topologies"""
    cluster_summary = nimbus_client.getClusterInfo()
    return cluster_summary.topologies


def list_topologies(env_name):
    """Prints out all running Storm topologies"""
    env_name, env_config = get_env_config(env_name)
    # Use ssh tunnel with Nimbus if use_ssh_for_nimbus is unspecified or True
    with ssh_tunnel(env_config) as (host, port):
        nimbus_client = get_nimbus_client(env_config, host=host, port=port)
        topologies = _list_topologies(nimbus_client)
    if not topologies:
        print('No topologies found.')
    else:
        columns = [field for field, default in
                   storm_thrift.TopologySummary.default_spec]
        print_stats_table('Topologies', topologies, columns, 'l')


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser('list',
                                      description=__doc__,
                                      help=main.__doc__)
    subparser.set_defaults(func=main)
    add_environment(subparser)


def main(args):
    """ List the currently running Storm topologies """
    list_topologies(args.environment)



"""
Create a JAR that can be used to deploy a topology to Storm.
"""

from __future__ import absolute_import, print_function, unicode_literals

import os
import shutil
import sys

from fabric.api import hide, local, settings

from ..util import prepare_topology
from .common import add_simple_jar


def jar_for_deploy(simple_jar=False):
    """ Build a jar to use for deploying the topology. """
    # Create _resources folder which will contain Python code in JAR
    prepare_topology()
    # Use Leiningen to clean up and build JAR
    jar_type = "JAR" if simple_jar else "Uber-JAR"
    print("Cleaning from prior builds...")
    sys.stdout.flush()
    with hide('running', 'stdout'):
        res = local("lein clean")
    if not res.succeeded:
        raise RuntimeError("Unable to run 'lein clean'!\nSTDOUT:\n{}"
                           "\nSTDERR:\n{}".format(res.stdout, res.stderr))
    print("Creating topology {}...".format(jar_type))
    sys.stdout.flush()
    cmd = "lein jar" if simple_jar else "lein uberjar"
    with hide('running'), settings(warn_only=True):
        res = local(cmd, capture=True)
        if not res.succeeded:
            raise RuntimeError("Unable to run '{}'!\nSTDOUT:\n{}"
                               "\nSTDERR:\n{}".format(cmd, res.stdout,
                                                      res.stderr))
        res = local(cmd, capture=True)
    # XXX: This will fail if more than one JAR is built
    lines = res.stdout.splitlines()
    for line in lines:
        line = line.strip()
        if not line.startswith("Created"):
            continue
        line = line.replace("Created ", "")
        # != is XOR
        if simple_jar != line.endswith("standalone.jar"):
            jar = line
            break
    else:
        raise RuntimeError("Failed to find JAR in '{}' output\STDOUT:\n{}"
                           "STDERR:\n{}".format(cmd, res.stdout, res.stderr))
    print("{} created: {}".format(jar_type, jar))
    sys.stdout.flush()
    print('Removing _resources temporary directory...', end='')
    sys.stdout.flush()
    resources_dir = os.path.join("_resources", "resources")
    if os.path.isdir(resources_dir):
        shutil.rmtree(resources_dir)
    print('done')
    return jar


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser('jar',
                                      description=__doc__,
                                      help=main.__doc__)
    subparser.set_defaults(func=main)
    add_simple_jar(subparser)


def main(args):
    """ Create a deployable JAR for a topology. """
    jar_for_deploy(simple_jar=args.simple_jar)



"""
Module to add streamparse-specific extensions to pystorm Spout class
"""

import pystorm

from ..dsl.spout import JavaSpoutSpec, ShellSpoutSpec
from .component import Component


class JavaSpout(Component):
    @classmethod
    def spec(cls, name=None, serialized_java=None, full_class_name=None,
             args_list=None, par=1, config=None, outputs=None):
        """Create a :class:`JavaSpoutSpec` for a Java Spout.

        This spec represents this Spout in a :class:`~streamparse.Topology`.

        You must add the appropriate entries to your classpath by editing your
        project's ``project.clj`` file in order for this to work.

        :param name:   Name of this Spout.  Defaults to name of
                       :class:`~streamparse.Topology` attribute this is assigned
                       to.
        :type name:    `str`
        :param serialized_java:  Serialized Java code representing the class.
                                 You must either specify this, or
                                 both ``full_class_name`` and ``args_list``.
        :type serialized_java: `bytes`
        :param full_class_name: Fully qualified class name (including the
                                package name)
        :type full_class_name: `str`
        :param args_list: A list of arguments to be passed to the constructor of
                          this class.
        :type args_list: `list` of basic data types
        :param par:    Parallelism hint for this Spout. See :ref:`parallelism`.
        :type par:     `int`
        :param config: Component-specific config settings to pass to Storm.
        :type config:  `dict`
        :param outputs: Outputs this JavaSpout will produce. Acceptable forms
                        are:

                        1.  A `list` of :class:`~streamparse.Stream` objects
                            describing the fields output on each stream.
                        2.  A `list` of `str` representing the fields output on
                            the ``default`` stream.
        """
        return JavaSpoutSpec(cls, name=name, serialized_java=serialized_java,
                             full_class_name=full_class_name,
                             args_list=args_list, par=par,
                             config=config, outputs=outputs)


class ShellSpout(Component):
    @classmethod
    def spec(cls, name=None, command=None, script=None, par=None, config=None,
             outputs=None):
        """Create a :class:`ShellSpoutSpec` for a non-Java, non-Python Spout.

        If you want to create a spec for a Python Spout, use
        :meth:`~streamparse.dsl.bolt.Spout.spec`.

        This spec represents this Spout in a :class:`~streamparse.Topology`.

        :param name:   Name of this Spout.  Defaults to name of
                       :class:`~streamparse.Topology` attribute this is assigned
                       to.
        :type name:    `str`
        :param command:  Path to command the Storm will execute.
        :type command: `str`
        :param script: Arguments to `command`.  Multiple arguments should just
                       be separated by spaces.
        :type script: `str`
        :param par:    Parallelism hint for this Spout.  For shell
                       Components, this works out to be the number of processes
                       running it in the the topology (across all machines).
                       See :ref:`parallelism`.
        :type par:     `int`
        :param config: Component-specific config settings to pass to Storm.
        :type config:  `dict`
        :param outputs: Outputs this ShellSpout will produce. Acceptable forms
                        are:

                        1.  A `list` of :class:`~streamparse.Stream` objects
                            describing the fields output on each stream.
                        2.  A `list` of `str` representing the fields output on
                            the ``default`` stream.
        """
        return ShellSpoutSpec(cls, command=command, script=script, name=name,
                              par=par, config=config, outputs=outputs)


class Spout(pystorm.spout.Spout, ShellSpout):
    """pystorm Spout with streamparse-specific additions"""
    @classmethod
    def spec(cls, name=None, par=None, config=None):
        """Create a :class:`~ShellBoltSpec` for a Python Spout.

        This spec represents this Spout in a :class:`~streamparse.Topology`.

        :param name:   Name of this Spout.  Defaults to name of
                       :class:`~streamparse.Topology` attribute this is assigned
                       to.
        :type name:    `str`
        :param par:    Parallelism hint for this Spout.  For Python
                       Components, this works out to be the number of Python
                       processes running it in the the topology (across all
                       machines).  See :ref:`parallelism`.

                       .. note::
                           This can also be specified as an attribute of your
                           :class:`~Spout` subclass.

        :type par:     `int`
        :param config: Component-specific config settings to pass to Storm.

                       .. note::
                           This can also be specified as an attribute of your
                           :class:`~Spout` subclass.

        :type config:  `dict`

        .. note::
            This method does not take a ``outputs`` argument because
            ``outputs`` should be an attribute of your :class:`~Spout` subclass.
        """
        return ShellSpoutSpec(cls, command='streamparse_run',
                              script='{}.{}'.format(cls.__module__,
                                                    cls.__name__),
                              name=name, par=par, config=config,
                              outputs=cls.outputs)



from __future__ import absolute_import, unicode_literals

import argparse
import unittest

from streamparse.cli.update_virtualenv import subparser_hook

from nose.tools import ok_


class UpdateVirtualenvTestCase(unittest.TestCase):

    def test_subparser_hook(self):
        parser = argparse.ArgumentParser()
        subparsers = parser.add_subparsers()
        subparser_hook(subparsers)

        subcommands = parser._optionals._actions[1].choices.keys()
        ok_('update_virtualenv' in subcommands)


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import, unicode_literals

import argparse
import unittest

from streamparse.cli import sparse

from nose.tools import ok_


class SparseTestCase(unittest.TestCase):

    def test_load_subparsers(self):
        parser = argparse.ArgumentParser()
        subparsers = parser.add_subparsers()
        sparse.load_subparsers(subparsers)
        # grab subcommands from subparsers
        subcommands = parser._optionals._actions[1].choices.keys()
        # we know quickstart will be a subcommand test others as needed
        ok_('quickstart' in subcommands)


if __name__ == '__main__':
    unittest.main()



"""
Component-level Specification

This module is called component to mirror organization of pystorm package.
"""
from __future__ import absolute_import

from copy import deepcopy

import simplejson as json
from six import iteritems, string_types

from ..thrift import storm_thrift
from .stream import Grouping, Stream
from .util import to_java_arg
from storm_thrift import (ComponentCommon, ComponentObject, GlobalStreamId,
                          JavaObject, ShellComponent, StreamInfo)


class ComponentSpec(object):
    """Describes the inputs, outputs, etc. for a Storm topology component.

    Generates some of the Thrift data structures that are needed to build a
    Storm topology.
    """
    def __init__(self, component_cls, name=None, inputs=None, par=1,
                 config=None, outputs=None):
        self.component_cls = component_cls
        self.name = name
        self.par = self._sanitize_par(component_cls, par)
        self.config = self._sanitize_config(component_cls, config)
        self.outputs = self._sanitize_outputs(component_cls, outputs)
        self.inputs = self._sanitize_inputs(inputs)
        self.common = ComponentCommon(inputs=self.inputs,
                                      streams=self.outputs,
                                      parallelism_hint=self.par,
                                      json_conf=self.config)

    @staticmethod
    def _sanitize_par(component_cls, par):
        """ Raises exceptions if `par` value is not a positive integer. """
        if par is None:
            par = component_cls.par
        if not isinstance(par, int):
            raise TypeError("Parallelism must be a integer greater than 0")
        elif par < 1:
            raise ValueError("Parallelism must be a integer greater than 0")
        return par

    @staticmethod
    def _sanitize_inputs(inputs):
        if isinstance(inputs, dict):
            for key, val in list(iteritems(inputs)):
                if not isinstance(key, GlobalStreamId):
                    if isinstance(key, ComponentSpec):
                        inputs[key['default']] = val
                        del inputs[key]
                    else:
                        raise TypeError('If inputs is a dict, it is expected to'
                                        ' map from GlobalStreamId or '
                                        'ComponentSpec objects to Grouping '
                                        'objects.  Given key: {!r}; Given '
                                        'value: {!r}'.format(key, val))
                if not isinstance(val, storm_thrift.Grouping):
                    raise TypeError('If inputs is a dict, it is expected to map'
                                    ' from GlobalStreamId or ComponentSpec '
                                    'objects to Grouping objects.  Given key: '
                                    '{!r}; Given value: {!r}'.format(key, val))
            input_dict = inputs
        else:
            if isinstance(inputs, ComponentSpec):
                inputs = [inputs]
            if isinstance(inputs, (list, tuple)):
                input_dict = {}
                for input_spec in inputs:
                    grouping = Grouping.SHUFFLE
                    if isinstance(input_spec, ComponentSpec):
                        component_id = input_spec.name or input_spec
                        stream_id = GlobalStreamId(componentId=component_id,
                                                   streamId='default')
                        # Can only automatically determine if grouping should be
                        # direct when given a ComponentSpec.  If
                        # GlobalStreamId, we're out of luck.
                        # TODO: Document this.
                        default_stream = input_spec.common.streams.get('default')
                        if default_stream is not None and default_stream.direct:
                            grouping = Grouping.DIRECT
                    elif isinstance(input_spec, GlobalStreamId):
                        stream_id = input_spec
                    else:
                        raise TypeError('Inputs must be ComponentSpec or '
                                        'GlobalStreamId objects.  Given: {!r}'
                                        .format(input_spec))
                    input_dict[stream_id] = grouping
            elif inputs is None:
                input_dict = {}
            else:
                raise TypeError('Inputs must either be a list, dict, or None.  '
                                'Given: {!r}'.format(inputs))
        return input_dict

    @staticmethod
    def _sanitize_config(component_cls, config):
        if config is None:
            config = component_cls.config
        if isinstance(config, dict):
            config = json.dumps(config)
        elif config is None:
            config = '{}'
        else:
            raise TypeError('Config must either be a dict or None.  Given: {!r}'
                            .format(config))
        return config

    @staticmethod
    def _sanitize_outputs(component_cls, outputs):
        if outputs is None:
            outputs = component_cls.outputs
        if outputs is None:
            outputs = []
        if isinstance(outputs, (list, tuple)):
            streams = {}
            for output in outputs:
                if isinstance(output, Stream):
                    streams[output.name] = StreamInfo(output_fields=output.fields,
                                                      direct=output.direct)
                # Strings are output fields for default stream
                elif isinstance(output, string_types):
                    default = streams.setdefault('default',
                                                 StreamInfo(output_fields=[],
                                                            direct=False))
                    default.output_fields.append(output)
                else:
                    raise TypeError('Outputs must either be a list of strings '
                                    'or a list of Streams.  Invalid entry: {!r}'
                                    .format(output))
        else:
            raise TypeError('Outputs must either be a list of strings or a list'
                            ' of Streams.  Given: {!r}'.format(outputs))
        return streams

    def __getitem__(self, stream):
        if stream not in self.common.streams:
            raise KeyError('Invalid stream for {}: {!r}. Valid streams are: '
                           '{}'.format(self.name, stream,
                                       list(self.common.streams.keys())))
        # If name is None, because it hasn't been set yet, use object instead
        component_id = self.name or self
        return GlobalStreamId(componentId=component_id, streamId=stream)

    def __repr__(self):
        """:returns: A string representation of the Specification. """
        attr_dict = deepcopy(self.__dict__)
        component_cls = attr_dict.pop('component_cls')
        repr_str = '{}({cls}'.format(self.__class__.__name__,
                                     cls=component_cls.__name__)
        for key, val in attr_dict.items():
            repr_str += ', {}={!r}'.format(key, val)
        repr_str += ')'
        return repr_str


class JavaComponentSpec(ComponentSpec):
    """ComponentSpec for JVM-based topology components."""
    def __init__(self, component_cls, name=None, serialized_java=None,
                 full_class_name=None, args_list=None, inputs=None,
                 par=None, config=None, outputs=None):
        super(JavaComponentSpec, self).__init__(component_cls,
                                                name=name, inputs=inputs,
                                                par=par, config=config,
                                                outputs=outputs)
        if serialized_java is not None:
            if isinstance(serialized_java, bytes):
                comp_object = ComponentObject(serialized_java=serialized_java)
                self.component_object = comp_object
            else:
                raise TypeError('serialized_java must be either bytes or None')
        else:
            if not full_class_name:
                raise ValueError('full_class_name is required')
            if args_list is None:
                raise TypeError('args_list must not be None')
            else:
                # Convert arguments to JavaObjectArgs
                for i, arg in enumerate(args_list):
                    args_list[i] = to_java_arg(arg)
            java_object = JavaObject(full_class_name=full_class_name,
                                     args_list=args_list)
            self.component_object = ComponentObject(java_object=java_object)


class ShellComponentSpec(ComponentSpec):
    """ComponentSpec for shell-based topology components (like Python ones)."""
    def __init__(self, component_cls, name=None, command=None, script=None,
                 inputs=None, par=1, config=None, outputs=None):
        super(ShellComponentSpec, self).__init__(component_cls,
                                                 name=name, inputs=inputs,
                                                 par=par, config=config,
                                                 outputs=outputs)
        if not command:
            raise ValueError('command is required')
        if script is None:
            raise TypeError('script must not be None.  If your command does not'
                            ' take arguments, specify the empty string for '
                            'script.')
        shell_component = ShellComponent(execution_command=command,
                                         script=script)
        self.component_object = ComponentObject(shell=shell_component)



# -*- coding: utf-8 -*-

# Copyright 2014-2015 Parsely, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
This module exists solely for version information so we only have to change it
in one place. Based on the suggestion `here. <http://bit.ly/16LbuJF>`_

:organization: Parsely
'''

def _safe_int(string):
    """ Simple function to convert strings into ints without dying. """
    try:
        return int(string)
    except ValueError:
        return string


__version__ = '3.1.1'
VERSION = tuple(_safe_int(x) for x in __version__.split('.'))



"""
Display slots used by every topology on the cluster
"""

from __future__ import absolute_import, print_function

from collections import Counter, defaultdict

from pkg_resources import parse_version
from prettytable import PrettyTable
from six import iteritems

from .common import add_environment
from ..util import get_ui_json, get_ui_jsons, storm_lib_version


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser('slot_usage',
                                      description=__doc__,
                                      help=main.__doc__)
    subparser.set_defaults(func=main)
    add_environment(subparser)


def display_slot_usage(env_name):
    print('Querying Storm UI REST service for slot usage stats (this can take a while)...')
    topology_summary_path = '/api/v1/topology/summary'
    topology_detail_path = '/api/v1/topology/{topology}'
    component_path = '/api/v1/topology/{topology}/component/{component}'
    topo_summary_json = get_ui_json(env_name, topology_summary_path)
    topology_ids = [x['id'] for x in topo_summary_json['topologies']]
    # Keep track of the number of workers used by each topology on each machine
    topology_worker_ports = defaultdict(lambda: defaultdict(set))
    topology_executor_counts = defaultdict(Counter)
    topology_names = set()
    topology_components = dict()
    topology_detail_jsons = get_ui_jsons(env_name,
                                         (topology_detail_path.format(topology=topology)
                                          for topology in topology_ids))

    for topology in topology_ids:
        topology_detail_json = topology_detail_jsons[topology_detail_path.format(topology=topology)]
        spouts = [x['spoutId'] for x in topology_detail_json['spouts']]
        bolts = [x['boltId'] for x in topology_detail_json['bolts']]
        topology_components[topology] = spouts + bolts

    comp_details = get_ui_jsons(env_name,
                                (component_path.format(topology=topology,
                                                       component=comp)
                                 for topology, comp_list in iteritems(topology_components)
                                 for comp in comp_list))

    for request_url, comp_detail in iteritems(comp_details):
        topology = request_url.split('/')[4]
        topology_detail_json = topology_detail_jsons[topology_detail_path.format(topology=topology)]
        for worker in comp_detail['executorStats']:
            topology_worker_ports[worker['host']][topology_detail_json['name']].add(worker['port'])
            topology_executor_counts[worker['host']][topology_detail_json['name']] += 1
            topology_names.add(topology_detail_json['name'])

    print("# Slot (and Executor) Counts by Topology")
    topology_names = sorted(topology_names)
    table = PrettyTable(["Host"] + topology_names)
    table.align = 'l'
    for host, host_dict in sorted(iteritems(topology_worker_ports)):
        row = [host] + ['{} ({})'.format(len(host_dict.get(topology, set())),
                                         topology_executor_counts[host][topology])
                        for topology in topology_names]
        table.add_row(row)
    print(table)
    print()


def main(args):
    """ Display slots used by every topology on the cluster. """
    storm_version = storm_lib_version()
    if storm_version >= parse_version('0.9.2-incubating'):
        display_slot_usage(args.environment)
    else:
        print("ERROR: Storm {0} does not support this command."
              .format(storm_version))



"""
Utilities for bootstrapping streamparse projects.
"""
from __future__ import absolute_import, print_function, unicode_literals

from contextlib import contextmanager
import os
import sys
import shutil

import pkg_resources
from fabric.colors import green, red, blue
from jinja2 import Environment, FileSystemLoader


_path_prefixes = []
_path_prefix = ''
_root = os.path.abspath(os.path.dirname(__file__))


def _here(*paths):
    path = os.path.join(*paths)
    filename = pkg_resources.resource_filename(__name__, path)
    return filename

_env = Environment(loader=FileSystemLoader(_here('project')))


@contextmanager
def _cd(path):
    global _path_prefixes
    global _path_prefix
    _path_prefixes.append(path)
    _path_prefix = '/'.join(_path_prefixes)
    yield
    _path_prefixes.pop()
    _path_prefix = '/'.join(_path_prefixes)


def _mkdir(path):
    path = '{}/{}'.format(_path_prefix, path) if _path_prefix != '' else path
    print('    {:<18} {}'.format(green('create'), path))
    os.makedirs(path)


def _cp(src, dest):
    dest = '{}/{}'.format(_path_prefix, dest) if _path_prefix != '' else dest
    print('    {:<18} {}'.format(green('create'), dest))
    shutil.copy(src, dest)


def _touch(filename):
    filename = '{}/{}'.format(_path_prefix, filename) if _path_prefix != '' \
               else filename
    print('    {:<18} {}'.format(green('create'), filename))
    with open(filename, 'w'):
        pass


def _generate(template_filename, dest):
    dest = '{}/{}'.format(_path_prefix, dest) if _path_prefix != '' \
           else dest
    print('    {:<18} {}'.format(green('create'), dest))
    template = _env.get_template(template_filename)
    with open(dest, 'w') as fp:
        fp.write(template.render())


def quickstart(project_name):
    # TODO: alternate way maybe to do all of this is do something like
    # glob.glob('project/**/*') and then we copy everything that's doesn't have
    # jinja2 in filename, generate the jinja2 stuff
    if os.path.exists(project_name):
        print('{}: folder "{}" already exists'.format(red('error'),
                                                      project_name))
        sys.exit(1)

    print('\nCreating your %s streamparse project...' % (blue(project_name)))
    _env.globals['project_name'] = project_name

    _mkdir(project_name)
    with _cd(project_name):
        _cp(_here('project', 'gitignore'), '.gitignore')
        _generate('config.jinja2.json', 'config.json')
        _cp(_here('project', 'fabfile.py'), 'fabfile.py')
        _generate('project.jinja2.clj', 'project.clj')
        _touch('README.md')
        _mkdir('src')
        with _cd('src'):
            _mkdir('bolts')
            with _cd('bolts'):
                _cp(_here('project', 'src', 'bolts', '__init__.py'),
                    '__init__.py')
                _cp(_here('project', 'src', 'bolts', 'wordcount.py'),
                    'wordcount.py')
            _mkdir('spouts')
            with _cd('spouts'):
                _cp(_here('project', 'src', 'spouts', '__init__.py'),
                    '__init__.py')
                _cp(_here('project', 'src', 'spouts', 'words.py'), 'words.py')
        _mkdir('topologies')
        with _cd('topologies'):
            _cp(_here('project', 'topologies', 'wordcount.py'), 'wordcount.py')
        _mkdir('virtualenvs')
        with _cd('virtualenvs'):
            _cp(_here('project', 'virtualenvs', 'wordcount.txt'),
                'wordcount.txt')

    print('Done.\n')
    print(('Try running your topology locally with:\n\n'
           '\tcd {}\n'
           '\tsparse run').format(project_name))



import os
from collections import Counter

from streamparse import Bolt


class WordCountBolt(Bolt):
    outputs = ['word', 'count']

    def initialize(self, conf, ctx):
        self.counter = Counter()
        self.pid = os.getpid()
        self.total = 0

    def _increment(self, word, inc_by):
        self.counter[word] += inc_by
        self.total += inc_by

    def process(self, tup):
        word = tup.values[0]
        self._increment(word, 10 if word == "dog" else 1)
        if self.total % 1000 == 0:
            self.logger.info("counted [{:,}] words [pid={}]".format(self.total,
                                                                    self.pid))
        self.emit([word, self.counter[word]])



"""
Create new streamparse project template.
"""

from __future__ import absolute_import

from streamparse.bootstrap import quickstart


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser('quickstart',
                                      description=__doc__,
                                      help=main.__doc__)
    subparser.set_defaults(func=main)
    subparser.add_argument('project_name',
                           help='Name of new streamparse project.')


def main(args):
    """ Create new streamparse project template. """
    quickstart(args.project_name)



from __future__ import absolute_import, unicode_literals

import argparse
import unittest

from nose.tools import ok_
try:
	from unittest.mock import patch
except ImportError:
	from mock import patch

from streamparse.cli.run import main, subparser_hook


class RunTestCase(unittest.TestCase):

    def test_subparser_hook(self):
        parser = argparse.ArgumentParser()
        subparsers = parser.add_subparsers()
        subparser_hook(subparsers)

        subcommands = parser._optionals._actions[1].choices.keys()
        ok_('run' in subcommands)


    @patch('streamparse.cli.run.run_local_topology', autospec=True)
    def test_main_args_passed(self, run_local_mock):
        parser = argparse.ArgumentParser()
        subparsers = parser.add_subparsers()
        subparser_hook(subparsers)

        args = parser.parse_args('run -e my_env -n my_topo --ackers 1'.split())

        main(args)
        run_local_mock.assert_called_with(name='my_topo',
                                          options={'topology.acker.executors': 1},
                                          env_name='my_env',
                                          time=0)


if __name__ == '__main__':
    unittest.main()



'''
This package makes it easier to work with Storm and Python.

:organization: Parsely
'''

from __future__ import absolute_import, print_function, unicode_literals

import logging

from . import bolt, cli, component, decorators, dsl, spout, storm
from .dsl import Grouping, Stream, Topology
from .storm import (BatchingBolt, Bolt, JavaBolt, JavaSpout, ShellBolt,
                    ShellSpout, Spout, StormHandler, TicklessBatchingBolt,
                    Tuple)
from .version import __version__, VERSION

# Enable default NullHandler to prevent "No handlers could be found for logger"
# https://lukasa.co.uk/2014/05/A_Brief_Digression_About_Logging/
logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    'BatchingBolt',
    'Bolt',
    'bolt',
    'cli',
    'component',
    'decorators',
    'dsl',
    'Grouping',
    'JavaBolt',
    'JavaSpout',
    'ShellBolt',
    'ShellSpout',
    'Spout',
    'spout',
    'storm',
    'StormHandler',
    'Stream',
    'TicklessBatchingBolt',
    'Topology',
    'Tuple'
]

__license__ = """
Copyright 2014-2015 Parsely, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""



"""
Submit a Storm topology to Nimbus.
"""

from __future__ import absolute_import, print_function, unicode_literals

import os
import sys
import time

import simplejson as json
from fabric.api import env
from six import itervalues

from ..dsl.component import JavaComponentSpec
from ..thrift import storm_thrift
from ..util import (activate_env, get_config, get_env_config, get_nimbus_client,
                    get_topology_definition, get_topology_from_file, ssh_tunnel)
from .common import (add_ackers, add_debug, add_environment, add_name,
                     add_options, add_override_name, add_wait, add_workers,
                     resolve_options)
from .jar import jar_for_deploy
from .kill import _kill_topology
from .list import _list_topologies
from .update_virtualenv import create_or_update_virtualenvs
from storm_thrift import ShellComponent


THRIFT_CHUNK_SIZE = 307200


def get_user_tasks():
    """Get tasks defined in a user's tasks.py and fabric.py file which is
    assumed to be in the current working directory.

    :returns: a `tuple` (invoke_tasks, fabric_tasks)
    """
    sys.path.insert(0, os.getcwd())
    try:
        import tasks as user_invoke
    except ImportError:
        user_invoke = None
    try:
        import fabfile as user_fabric
    except ImportError:
        user_fabric = None
    return user_invoke, user_fabric


def is_safe_to_submit(topology_name, nimbus_client):
    """Is topology not in list of current topologies?"""
    topologies = _list_topologies(nimbus_client)
    safe = not any(topology.name == topology_name for topology in topologies)
    return safe


def _kill_existing_topology(topology_name, force, wait, nimbus_client):
    if force and not is_safe_to_submit(topology_name, nimbus_client):
        print("Killing current \"{}\" topology.".format(topology_name))
        sys.stdout.flush()
        _kill_topology(topology_name, nimbus_client, wait=wait)
        while not is_safe_to_submit(topology_name, nimbus_client):
            print("Waiting for topology {} to quit...".format(topology_name))
            sys.stdout.flush()
            time.sleep(0.5)
        print("Killed.")
        sys.stdout.flush()


def _submit_topology(topology_name, topology_class, uploaded_jar, config,
                     env_config, nimbus_client, options=None):
    if options.get('pystorm.log.path'):
        print("Routing Python logging to {}.".format(options['pystorm.log.path']))
        sys.stdout.flush()

    serializer = env_config.get('serializer', config.get('serializer', None))
    if serializer is not None:
        # Set serializer arg in bolts
        for thrift_bolt in itervalues(topology_class.thrift_bolts):
            inner_shell = thrift_bolt.bolt_object.shell
            if inner_shell is not None:
                inner_shell.script = '-s {} {}'.format(serializer,
                                                       inner_shell.script)
        # Set serializer arg in spouts
        for thrift_spout in itervalues(topology_class.thrift_spouts):
            inner_shell = thrift_spout.spout_object.shell
            if inner_shell is not None:
                inner_shell.script = '-s {} {}'.format(serializer,
                                                       inner_shell.script)

    print("Submitting {} topology to nimbus...".format(topology_name), end='')
    sys.stdout.flush()
    nimbus_client.submitTopology(name=topology_name,
                                 uploadedJarLocation=uploaded_jar,
                                 jsonConf=json.dumps(options),
                                 topology=topology_class.thrift_topology)
    print('done')


def _pre_submit_hooks(topology_name, env_name, env_config):
    """Pre-submit hooks for invoke and fabric."""
    user_invoke, user_fabric = get_user_tasks()
    pre_submit_invoke = getattr(user_invoke, "pre_submit", None)
    if callable(pre_submit_invoke):
        pre_submit_invoke(topology_name, env_name, env_config)
    pre_submit_fabric = getattr(user_fabric, "pre_submit", None)
    if callable(pre_submit_fabric):
        pre_submit_fabric(topology_name, env_name, env_config)


def _post_submit_hooks(topology_name, env_name, env_config):
    """Post-submit hooks for invoke and fabric."""
    user_invoke, user_fabric = get_user_tasks()
    post_submit_invoke = getattr(user_invoke, "post_submit", None)
    if callable(post_submit_invoke):
        post_submit_invoke(topology_name, env_name, env_config)
    post_submit_fabric = getattr(user_fabric, "post_submit", None)
    if callable(post_submit_fabric):
        post_submit_fabric(topology_name, env_name, env_config)


def _upload_jar(nimbus_client, local_path):
    upload_location = nimbus_client.beginFileUpload()
    print("Uploading topology jar {} to assigned location: {}"
          .format(local_path, upload_location))
    total_bytes = os.path.getsize(local_path)
    bytes_uploaded = 0
    with open(local_path, 'rb') as local_jar:
        while True:
            print("Uploaded {}/{} bytes".format(bytes_uploaded, total_bytes),
                  end='\r')
            sys.stdout.flush()
            curr_chunk = local_jar.read(THRIFT_CHUNK_SIZE)
            if not curr_chunk:
                break
            nimbus_client.uploadChunk(upload_location, curr_chunk)
            bytes_uploaded += len(curr_chunk)
        nimbus_client.finishFileUpload(upload_location)
        print("Uploaded {}/{} bytes".format(bytes_uploaded, total_bytes))
        sys.stdout.flush()
    return upload_location


def submit_topology(name=None, env_name=None, options=None, force=False,
                    wait=None, simple_jar=True, override_name=None):
    """Submit a topology to a remote Storm cluster."""
    config = get_config()
    name, topology_file = get_topology_definition(name)
    env_name, env_config = get_env_config(env_name)
    topology_class = get_topology_from_file(topology_file)
    if override_name is None:
        override_name = name

    # Check if we need to maintain virtualenv during the process
    use_venv = env_config.get('use_virtualenv', True)

    # Check if user wants to install virtualenv during the process
    install_venv = env_config.get('install_virtualenv', use_venv)

    # Setup the fabric env dictionary
    activate_env(env_name)
    # Run pre_submit actions provided by project
    _pre_submit_hooks(name, env_name, env_config)

    # If using virtualenv, set it up, and make sure paths are correct in specs
    if use_venv:
        if install_venv:
            create_or_update_virtualenvs(env_name, name,
                                         override_name=override_name)
        streamparse_run_path = '/'.join([env.virtualenv_root, override_name,
                                         'bin', 'streamparse_run'])
        # Update python paths in bolts
        for thrift_bolt in itervalues(topology_class.thrift_bolts):
            inner_shell = thrift_bolt.bolt_object.shell
            if isinstance(inner_shell, ShellComponent):
                if 'streamparse_run' in inner_shell.execution_command:
                    inner_shell.execution_command = streamparse_run_path
        # Update python paths in spouts
        for thrift_spout in itervalues(topology_class.thrift_spouts):
            inner_shell = thrift_spout.spout_object.shell
            if isinstance(inner_shell, ShellComponent):
                if 'streamparse_run' in inner_shell.execution_command:
                    inner_shell.execution_command = streamparse_run_path

    # Handle option conflicts
    options = resolve_options(options, env_config, topology_class,
                              override_name)

    # Check topology for JVM stuff to see if we need to create uber-jar
    if simple_jar:
        simple_jar = not any(isinstance(spec, JavaComponentSpec)
                             for spec in topology_class.specs)

    # Prepare a JAR that doesn't have Storm dependencies packaged
    topology_jar = jar_for_deploy(simple_jar=simple_jar)

    if name != override_name:
        print('Deploying "{}" topology with name "{}"...'.format(name,
                                                                 override_name))
    else:
        print('Deploying "{}" topology...'.format(name))
    sys.stdout.flush()
    # Use ssh tunnel with Nimbus if use_ssh_for_nimbus is unspecified or True
    with ssh_tunnel(env_config) as (host, port):
        nimbus_client = get_nimbus_client(env_config, host=host, port=port)
        uploaded_jar = _upload_jar(nimbus_client, topology_jar)
        _kill_existing_topology(override_name, force, wait, nimbus_client)
        _submit_topology(override_name, topology_class, uploaded_jar, config,
                         env_config, nimbus_client, options=options)
    _post_submit_hooks(name, env_name, env_config)


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser('submit',
                                      description=__doc__,
                                      help=main.__doc__)
    subparser.set_defaults(func=main)
    add_ackers(subparser)
    add_debug(subparser)
    add_environment(subparser)
    subparser.add_argument('-f', '--force',
                           action='store_true',
                           help='Force a topology to submit by killing any '
                                'currently running topologies with the same '
                                'name.')
    add_name(subparser)
    add_options(subparser)
    add_override_name(subparser)
    subparser.add_argument('-u', '--uber_jar',
                           help='Build an Uber-JAR even if you have no Java '
                                'components in your topology.  Useful if you '
                                'are providing your own seriailzer class.',
                           dest='simple_jar', action='store_false')
    add_wait(subparser)
    add_workers(subparser)


def main(args):
    """ Submit a Storm topology to Nimbus. """
    submit_topology(name=args.name, env_name=args.environment,
                    options=args.options, force=args.force, wait=args.wait,
                    simple_jar=args.simple_jar,
                    override_name=args.override_name)



"""
Visualize a topology using Graphviz.
"""
# This code is based on the code from Dask's dot module
from __future__ import absolute_import, print_function

import os
from functools import partial

from six import iteritems

from ..dsl.spout import JavaSpoutSpec, ShellSpoutSpec
from ..util import get_topology_definition, get_topology_from_file
from .common import add_name

try:
    import graphviz
    HAVE_GRAPHVIZ = True
except:
    HAVE_GRAPHVIZ = False


IPYTHON_IMAGE_FORMATS = frozenset(['jpeg', 'png'])
IPYTHON_NO_DISPLAY_FORMATS = frozenset(['dot', 'pdf'])


def to_graphviz(topology_class, node_attr=None, edge_attr=None, **kwargs):
    """Convert a Topology into a DiGraph"""
    if not HAVE_GRAPHVIZ:
        raise ImportError('The visualize command requires the `graphviz` Python'
                          ' library and `graphviz` system library to be '
                          'installed.')
    attributes = {'fontsize': '16',
                  'fontcolor': 'white',
                  'bgcolor': '#333333',
                  'rankdir': 'LR'}
    node_attributes = {'fontname': 'Helvetica',
                       'fontcolor': 'white',
                       'color': 'white',
                       'style': 'filled',
                       'fillcolor': '#006699'}
    edge_attributes = {'style': 'solid',
                       'color': 'white',
                       'arrowhead': 'open',
                       'fontname': 'Helvetica',
                       'fontsize': '12',
                       'fontcolor': 'white'}
    attributes.update(kwargs)
    if node_attr is not None:
        node_attributes.update(node_attr)
    if edge_attr is not None:
        edge_attributes.update(edge_attr)
    g = graphviz.Digraph(graph_attr=attributes, node_attr=node_attributes,
                         edge_attr=edge_attributes)

    all_specs = {}
    all_specs.update(topology_class.thrift_bolts)
    all_specs.update(topology_class.thrift_spouts)

    sametail_nodes = set()

    for spec in topology_class.specs:
        if isinstance(spec, (JavaSpoutSpec, ShellSpoutSpec)):
            shape = 'box'
        else:
            shape = None
        g.node(spec.name, label=spec.name, shape=shape)
        for stream_id, grouping in list(iteritems(spec.inputs)):
            parent = stream_id.componentId
            outputs = all_specs[parent].common.streams[stream_id.streamId].output_fields
            label = ('Stream: {}\lFields: {}\lGrouping: {}\l'
                     .format(stream_id.streamId, outputs, grouping))
            sametail = '{}-{}'.format(parent, stream_id.streamId)
            if sametail not in sametail_nodes:
                g.node(sametail, shape='point', width='0')
                g.edge(parent, sametail, label=label, dir='none')
                sametail_nodes.add(sametail)
            g.edge(sametail, spec.name, samehead=str(outputs))

    return g


def _get_display_cls(format):
    """
    Get the appropriate IPython display class for `format`.

    Returns `IPython.display.SVG` if format=='svg', otherwise
    `IPython.display.Image`.

    If IPython is not importable, return dummy function that swallows its
    arguments and returns None.
    """
    dummy = lambda *args, **kwargs: None
    try:
        import IPython.display as display
    except ImportError:
        # Can't return a display object if no IPython.
        return dummy

    if format in IPYTHON_NO_DISPLAY_FORMATS:
        # IPython can't display this format natively, so just return None.
        return dummy
    elif format in IPYTHON_IMAGE_FORMATS:
        # Partially apply `format` so that `Image` and `SVG` supply a uniform
        # interface to the caller.
        return partial(display.Image, format=format)
    elif format == 'svg':
        return display.SVG
    else:
        raise ValueError("Unknown format '%s' passed to `dot_graph`" % format)


def visualize_topology(name=None, filename=None, format=None, **kwargs):
    """
    Render a topology graph using dot.

    If `filename` is not None, write a file to disk with that name in the
    format specified by `format`.  `filename` should not include an extension.

    Parameters
    ----------
    name : str
        The name of the topology to display.
    filename : str or None, optional
        The name (without an extension) of the file to write to disk.  If
        `filename` is None, no file will be written, and we communicate with
        dot using only pipes.  Default is `name`.
    format : {'png', 'pdf', 'dot', 'svg', 'jpeg', 'jpg'}, optional
        Format in which to write output file.  Default is 'png'.
    **kwargs
        Additional keyword arguments to forward to `to_graphviz`.

    Returns
    -------
    result : None or IPython.display.Image or IPython.display.SVG  (See below.)

    Notes
    -----
    If IPython is installed, we return an IPython.display object in the
    requested format.  If IPython is not installed, we just return None.

    We always return None if format is 'pdf' or 'dot', because IPython can't
    display these formats natively. Passing these formats with filename=None
    will not produce any useful output.
    """
    name, topology_file = get_topology_definition(name)
    topology_class = get_topology_from_file(topology_file)

    if filename is None:
        filename = name

    g = to_graphviz(topology_class, **kwargs)

    fmts = ['.png', '.pdf', '.dot', '.svg', '.jpeg', '.jpg']
    if format is None and any(filename.lower().endswith(fmt) for fmt in fmts):
        filename, format = os.path.splitext(filename)
        format = format[1:].lower()

    if format is None:
        format = 'png'

    data = g.pipe(format=format)
    if not data:
        raise RuntimeError("Graphviz failed to properly produce an image. "
                           "This probably means your installation of graphviz "
                           "is missing png support. See: "
                           "https://github.com/ContinuumIO/anaconda-issues/"
                           "issues/485 for more information.")

    display_cls = _get_display_cls(format)

    if not filename:
        return display_cls(data=data)

    full_filename = '.'.join([filename, format])
    with open(full_filename, 'wb') as f:
        f.write(data)

    return display_cls(filename=full_filename)


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser('visualize',
                                      description=__doc__,
                                      help=main.__doc__)
    subparser.set_defaults(func=main)
    add_name(subparser)
    subparser.add_argument('-f', '--format',
                           help='File extension for graph file. Defaults to PNG')
    subparser.add_argument('-o', '--output_file',
                           help='Name of output file. Defaults to NAME.FORMAT')


def main(args):
    """Create a Graphviz visualization of the topology"""
    visualize_topology(name=args.name, format=args.format,
                       filename=args.output_file)



def pre_submit(topology_name, env_name, env_config):
    """Override this function to perform custom actions prior to topology
    submission. No SSH tunnels will be active when this function is called."""
    pass


def post_submit(topo_name, env_name, env_config):
    """Override this function to perform custom actions after topology
    submission. Note that the SSH tunnel to Nimbus will still be active
    when this function is called."""
    pass



"""
Helper script for starting up bolts and spouts.
"""

import argparse
import importlib
import os
import sys

from pystorm.component import _SERIALIZERS


def main():
    """main entry point for Python bolts and spouts"""
    parser = argparse.ArgumentParser(description='Run a bolt/spout class',
                                     epilog='This is internal to streamparse '
                                            'and is used to run spout and bolt '
                                            'classes via ``python -m '
                                            'streamparse.run <class name>``.')
    parser.add_argument('target_class', help='The bolt/spout class to start.')
    parser.add_argument('-s', '--serializer',
                        help='The serialization protocol to use to talk to '
                             'Storm.',
                        choices=_SERIALIZERS.keys(),
                        default='json')
    # Storm sends everything as one string, which is not great
    if len(sys.argv) == 2:
        sys.argv = [sys.argv[0]] + sys.argv[1].split()
    args = parser.parse_args()
    # Add current directory to sys.path so imports will work
    sys.path.append(os.getcwd())
    # Import module
    mod_name, cls_name = args.target_class.rsplit('.', 1)
    mod = importlib.import_module(mod_name)
    # Get class from module and run it
    cls = getattr(mod, cls_name)
    cls(serializer=args.serializer).run()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Copyright 2014-2015 Parsely, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import re
import sys

from setuptools import setup, find_packages

# Get version without importing, which avoids dependency issues
def get_version():
    with open('streamparse/version.py') as version_file:
        return re.search(r"""__version__\s+=\s+(['"])(?P<version>.+?)\1""",
                         version_file.read()).group('version')

def readme():
    ''' Returns README.rst contents as str '''
    with open('README.rst') as f:
        return f.read()


install_requires = [
    'fabric3',
    'jinja2',
    'requests',
    'prettytable',
    'six>=1.5',
    'simplejson',
    'pystorm>=3.0.3',
    'thriftpy>=0.3.2',
    'ruamel.yaml'
]

if sys.version_info.major < 3:
    install_requires.append('contextlib2')

lint_requires = [
    'pep8',
    'pyflakes'
]

if sys.version_info.major < 3:
    tests_require = ['graphviz', 'mock', 'nose', 'unittest2']
else:
    tests_require = ['graphviz', 'mock', 'nose']

dependency_links = []
setup_requires = []
if 'nosetests' in sys.argv[1:]:
    setup_requires.append('nose')

setup(
    name='streamparse',
    version=get_version(),
    author='Parsely, Inc.',
    author_email='hello@parsely.com',
    url='https://github.com/Parsely/streamparse',
    description=('streamparse lets you run Python code against real-time '
                 'streams of data. Integrates with Apache Storm.'),
    long_description=readme(),
    license='Apache License 2.0',
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'sparse = streamparse.cli.sparse:main',
            'streamparse = streamparse.cli.sparse:main',
            'streamparse_run = streamparse.run:main'
        ]
    },
    install_requires=install_requires,
    tests_require=tests_require,
    setup_requires=setup_requires,
    extras_require={
        'test': tests_require,
        'all': install_requires + tests_require,
        'docs': ['sphinx'] + tests_require,
        'lint': lint_requires
    },
    dependency_links=dependency_links,
    zip_safe=False,
    test_suite='nose.collector',
    include_package_data=True,
)



"""
Topology base class
"""
from __future__ import absolute_import

from copy import deepcopy

import simplejson as json
from pystorm.component import Component
from six import add_metaclass, iteritems, itervalues, string_types
from thriftpy.transport import TMemoryBuffer
from thriftpy.protocol import TBinaryProtocol

from ..thrift import storm_thrift
from .bolt import JavaBoltSpec, ShellBoltSpec
from .component import ComponentSpec, ShellComponentSpec
from .spout import JavaSpoutSpec, ShellSpoutSpec
from .util import to_python_arg_list


class TopologyType(type):
    """Class to define a Storm topology in a Python DSL."""

    def __new__(mcs, classname, bases, class_dict):
        bolt_specs = {}
        spout_specs = {}
        # Copy ComponentSpec items out of class_dict
        specs = TopologyType.class_dict_to_specs(class_dict)
        # Perform checks
        for spec in itervalues(specs):
            if isinstance(spec, (JavaBoltSpec, ShellBoltSpec)):
                TopologyType.add_bolt_spec(spec, bolt_specs)
            elif isinstance(spec, (JavaSpoutSpec, ShellSpoutSpec)):
                TopologyType.add_spout_spec(spec, spout_specs)
            else:
                raise TypeError('Specifications should either be bolts or '
                                'spouts.  Given: {!r}'.format(spec))
            TopologyType.clean_spec_inputs(spec, specs)
        if classname != 'Topology' and not spout_specs:
            raise ValueError('A Topology requires at least one Spout')
        if 'config' in class_dict:
            config_dict = class_dict['config']
            if not isinstance(config_dict, dict):
                raise TypeError('Topology config must be a dictionary. Given: '
                                '{!r}'.format(config_dict))
        else:
            class_dict['config'] = {}
        class_dict['thrift_bolts'] = bolt_specs
        class_dict['thrift_spouts'] = spout_specs
        class_dict['specs'] = list(specs.values())
        class_dict['thrift_topology'] = storm_thrift.StormTopology(spouts=spout_specs,
                                                                   bolts=bolt_specs,
                                                                   state_spouts={})
        return type.__new__(mcs, classname, bases, class_dict)

    @classmethod
    def class_dict_to_specs(mcs, class_dict):
        """Extract valid `ComponentSpec` entries from `Topology.__dict__`."""
        specs = {}
        # Set spec names first
        for name, spec in iteritems(class_dict):
            if isinstance(spec, ComponentSpec):
                # Use the variable name as the specification name.
                if spec.name is None:
                    spec.name = name
                if spec.name in specs:
                    raise ValueError("Duplicate component name: {}"
                                     .format(spec.name))
                else:
                    specs[spec.name] = spec
            elif isinstance(spec, Component):
                raise TypeError('Topology classes should have ComponentSpec '
                                'attributes.  Did you forget to call the spec '
                                'class method for your component?  Given: {!r}'
                                .format(spec))
        return specs

    @classmethod
    def add_bolt_spec(mcs, spec, bolt_specs):
        """Add valid Bolt specs to `bolt_specs`; raise exceptions for others."""
        if not spec.inputs:
            cls_name = spec.component_cls.__name__
            raise ValueError('{} "{}" requires at least one input, because it '
                             'is a Bolt.'.format(cls_name, spec.name))
        bolt_specs[spec.name] = storm_thrift.Bolt(bolt_object=spec.component_object,
                                                  common=spec.common)

    @classmethod
    def add_spout_spec(mcs, spec, spout_specs):
        """Add valid Spout specs to `spout_specs`; raise exceptions for others.
        """
        if not spec.outputs:
            cls_name = spec.component_cls.__name__
            raise ValueError('{} "{}" requires at least one output, because it '
                             'is a Spout'.format(cls_name, spec.name))
        spout_specs[spec.name] = storm_thrift.SpoutSpec(spout_object=spec.component_object,
                                                        common=spec.common)

    @classmethod
    def clean_spec_inputs(mcs, spec, specs):
        """Convert `spec.inputs` to a dict mapping from stream IDs to groupings.
        """
        if spec.inputs is None:
            spec.inputs = {}
        for stream_id, grouping in list(iteritems(spec.inputs)):
            if isinstance(stream_id.componentId, ComponentSpec):
                # Have to reinsert key after fix because hash changes
                del spec.inputs[stream_id]
                stream_id.componentId = stream_id.componentId.name
                spec.inputs[stream_id] = grouping
            # This should never happen, but it's worth checking for
            elif stream_id.componentId is None:
                raise TypeError('GlobalStreamId.componentId cannot be None.')
            # Check for invalid fields grouping
            stream_comp = specs[stream_id.componentId]
            valid_fields = set(stream_comp.outputs[stream_id.streamId]
                               .output_fields)
            if grouping.fields is not None:
                for field in grouping.fields:
                    if field not in valid_fields:
                        raise ValueError('Field {!r} specified in grouping is '
                                         'not a valid output field for the {!r}'
                                         ' {!r} stream.'.format(field,
                                                                stream_comp.name,
                                                                stream_id.streamId))

    def __repr__(cls):
        """:returns: A string representation of the topology"""
        return repr(getattr(cls, '_topology', None))


@add_metaclass(TopologyType)
class Topology(object):
    """Class to define a Storm topology in a Python DSL."""
    @classmethod
    def write(cls, stream):
        """Write the topology to a stream or file.

        Typically used to write to Nimbus.

        .. note::
            This will not save the `specs` attribute, as that is not part of
            the Thrift output.
        """
        def write_it(stream):
            transport_out = TMemoryBuffer()
            protocol_out = TBinaryProtocol(transport_out)
            cls._topology.write(protocol_out)
            transport_bytes = transport_out.getvalue()
            stream.write(transport_bytes)

        if isinstance(stream, string_types):
            with open(stream, 'wb') as output_file:
                write_it(output_file)
        else:
            write_it(stream)

    @classmethod
    def read(cls, stream):
        """Read a topology from a stream or file.

        .. note::
            This will not properly reconstruct the `specs` attribute, as that is
            not included in the Thrift output.
        """
        def read_it(stream):
            stream_bytes = stream.read()
            transport_in = TMemoryBuffer(stream_bytes)
            protocol_in = TBinaryProtocol(transport_in)
            topology = storm_thrift.StormTopology()
            topology.read(protocol_in)
            cls._topology = topology
            cls.thrift_bolts = topology.bolts
            cls.thrift_spouts = topology.spouts
            # Can't reconstruct Python specs from Thrift.
            cls.specs = []

        if isinstance(stream, string_types):
            with open(stream, 'rb') as input_file:
                return read_it(input_file)
        else:
            return read_it(stream)

    @staticmethod
    def _spec_to_flux_dict(spec):
        """Convert a ComponentSpec into a dict as expected by Flux"""
        flux_dict = {'id': spec.name,
                     'constructorArgs': []}
        if isinstance(spec, ShellComponentSpec):
            if isinstance(spec, ShellBoltSpec):
                flux_dict['className'] = 'org.apache.storm.flux.wrappers.bolts.FluxShellBolt'
            else:
                flux_dict['className'] = 'org.apache.storm.flux.wrappers.spouts.FluxShellSpout'
            shell_object = spec.component_object.shell
            flux_dict['constructorArgs'].append([shell_object.execution_command,
                                                 shell_object.script])
            if not spec.outputs:
                flux_dict['constructorArgs'].append(['NONE_BUT_FLUX_WANTS_SOMETHING_HERE'])
            for output_stream in spec.outputs.keys():
                if output_stream == 'default':
                    output_fields = spec.outputs['default'].output_fields
                    flux_dict['constructorArgs'].append(output_fields)
                else:
                    if 'configMethods' not in flux_dict:
                        flux_dict['configMethods'] = []
                    flux_dict['configMethods'].append({
                        'name': 'setNamedStream',
                        'args': [
                            output_stream,
                            spec.outputs[output_stream].output_fields
                        ]
                    })
        else:
            if spec.component_object.serialized_java is not None:
                raise TypeError('Flux does not support specifying serialized '
                                'Java objects.  Given: {!r}'.format(spec))
            java_object = spec.component_object.java_object
            flux_dict['className'] = java_object.full_class_name
            # Convert JavaObjectArg instances into basic data types
            flux_dict['constructorArgs'] = to_python_arg_list(java_object.args_list)
        return flux_dict

    @staticmethod
    def _stream_to_flux_dict(spec, global_stream, grouping):
        """Convert a GlobalStreamId into a dict as expected by Flux"""
        flux_dict = {'from': global_stream.componentId,
                     'to': spec.name}
        grouping_dict = {'streamId': global_stream.streamId}
        for key, val in grouping.__dict__.items():
            if val is not None:
                grouping_dict['type'] = key.upper()
                if key == 'fields':
                    if val:
                        grouping_dict['args'] = val
                    else:
                        grouping_dict['type'] = 'GLOBAL'
                elif key == 'custom_object':
                    grouping_dict['type'] = 'CUSTOM'
                    class_dict = {'className': val.full_class_name,
                                  'args': to_python_arg_list(val.arg_list)}
                    grouping_dict['customClass'] = class_dict
        flux_dict['grouping'] = grouping_dict
        return flux_dict

    @classmethod
    def to_flux_dict(cls, name):
        """Convert topology to dict that can written out as Flux YAML file."""
        flux_dict = {'name': name,
                     'bolts': [],
                     'spouts': [],
                     'streams': []}
        for spec in cls.specs:
            if isinstance(spec, (JavaBoltSpec, ShellBoltSpec)):
                flux_dict['bolts'].append(cls._spec_to_flux_dict(spec))
                for global_stream, grouping in spec.inputs.items():
                    stream_dict = cls._stream_to_flux_dict(spec,
                                                           global_stream,
                                                           grouping)
                    flux_dict['streams'].append(stream_dict)
            elif isinstance(spec, (JavaSpoutSpec, ShellSpoutSpec)):
                flux_dict['spouts'].append(cls._spec_to_flux_dict(spec))
            else:
                raise TypeError('Specifications should either be bolts or '
                                'spouts.  Given: {!r}'.format(spec))
        flux_dict = {key: val for key, val in flux_dict.items() if val}
        return flux_dict



"""
Kill the specified Storm topology.
"""

from __future__ import absolute_import

from ..thrift import storm_thrift
from ..util import (get_topology_definition, get_env_config, get_nimbus_client,
                    ssh_tunnel)
from .common import add_environment, add_name, add_override_name, add_wait


def _kill_topology(topology_name, nimbus_client, wait=None):
    kill_opts = storm_thrift.KillOptions(wait_secs=wait)
    nimbus_client.killTopologyWithOpts(name=topology_name, options=kill_opts)


def kill_topology(topology_name=None, env_name=None, wait=None,
                  override_name=None):
    topology_name = get_topology_definition(topology_name)[0]
    env_name, env_config = get_env_config(env_name)
    # Use ssh tunnel with Nimbus if use_ssh_for_nimbus is unspecified or True
    with ssh_tunnel(env_config) as (host, port):
        nimbus_client = get_nimbus_client(env_config, host=host, port=port)
        return _kill_topology(override_name or topology_name, nimbus_client,
                              wait=wait)


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser('kill',
                                      description=__doc__,
                                      help=main.__doc__)
    subparser.set_defaults(func=main)
    add_environment(subparser)
    add_name(subparser)
    add_override_name(subparser)
    add_wait(subparser)


def main(args):
    """ Kill the specified Storm topology """
    kill_topology(topology_name=args.name, env_name=args.environment,
                  wait=args.wait, override_name=None)



"""
Streams and Groupings
"""
from __future__ import absolute_import

from six import iteritems, string_types

from ..thrift import storm_thrift
from .util import to_java_arg
from storm_thrift import NullStruct


class Stream(storm_thrift.StreamInfo):
    """
    A Storm output stream
    """
    def __init__(self, fields=None, name='default', direct=False):
        """
        :param fields: Field names for this stream.
        :type fields:  `list` or `tuple` of `str`
        :param name:   Name of stream.  Defaults to ``default``.
        :type name:    `str`
        :param direct: Whether or not this stream is direct.  Default is `False`.
                       See :attr:`~streamparse.dsl.stream.Grouping.DIRECT`.
        :type direct:  `bool`
        """
        if fields is None:
            fields = []
        elif isinstance(fields, (list, tuple)):
            fields = list(fields)
            for field in fields:
                if not isinstance(field, string_types):
                    raise TypeError('All field names must be strings; given: '
                                    '{!r}'.format(field))
        else:
            raise TypeError('Stream fields must be a list, tuple, or None; '
                            'given: {!r}'.format(fields))
        self.fields = fields
        if isinstance(name, string_types):
            self.name = name
        else:
            raise TypeError('Stream name must be a string; given: {!r}'
                            .format(name))
        if isinstance(direct, bool):
            self.direct = direct
        else:
            raise TypeError('"direct" must be either True or False; given: {!r}'
                            .format(direct))


class _Grouping(storm_thrift.Grouping):
    """
    Version of `storm_thrift.Grouping` that has better __str__.
    """
    def __repr__(self):
        for name, val in iteritems(vars(self)):
            if not name.startswith('_') and val is not None:
                if isinstance(val, NullStruct):
                    return '{}'.format(name.upper())
                else:
                    return '{}({!r})'.format(name, val)


class Grouping(object):
    """
    A Grouping describes how Tuples should be distributed to the tasks of a
    Bolt listening on a particular stream.

    When no Grouping is specified, it defaults to `SHUFFLE` for normal streams,
    and `DIRECT` for direct streams.

    :ivar SHUFFLE: Tuples are randomly distributed across the Bolt's tasks in a
                   way such that each Bolt is guaranteed to get an equal number
                   of Tuples.
    :ivar GLOBAL: The entire stream goes to a single one of the Bolt's tasks.
                  Specifically, it goes to the task with the lowest id.
    :ivar DIRECT: This is a special kind of grouping. A stream grouped this way
                  means that the producer of the Tuple decides which task of the
                  consumer will receive this Tuple. Direct groupings can only be
                  declared on streams that have been declared as direct streams.
                  Tuples emitted to a direct stream must be emitted using the
                  the `direct_task` parameter to the
                  :meth:`streamparse.Bolt.emit` and
                  :meth:`streamparse.Spout.emit` methods.
    :ivar ALL: The stream is replicated across all the Bolt's tasks. Use this
               grouping with care.
    :ivar NONE: This grouping specifies that you don't care how the stream is
                grouped. Currently, none groupings are equivalent to shuffle
                groupings. Eventually though, Storm will push down Bolts with
                none groupings to execute in the same thread as the Bolt or
                Spout they subscribe from (when possible).
    :ivar LOCAL_OR_SHUFFLE: If the target Bolt has one or more tasks in the
                            same worker process, Tuples will be shuffled to
                            just those in-process tasks. Otherwise, this acts
                            like a normal shuffle grouping.
    """
    __slots__ = ()

    SHUFFLE = _Grouping(shuffle=NullStruct())
    GLOBAL = _Grouping(fields=[])
    DIRECT = _Grouping(direct=NullStruct())
    ALL = _Grouping(all=NullStruct())
    NONE = _Grouping(none=NullStruct())
    LOCAL_OR_SHUFFLE = _Grouping(local_or_shuffle=NullStruct())

    @classmethod
    def fields(cls, *fields):
        """The stream is partitioned by the fields specified in the grouping.

        For example, if the stream is grouped by the `user-id` field, Tuples
        with the same `user-id` will always go to the same task, but Tuples with
        different `user-id`'s may go to different tasks."""
        if len(fields) == 1 and isinstance(fields[0], list):
            fields = fields[0]
        else:
            fields = list(fields)
        if not fields:
            raise ValueError('List cannot be empty for fields grouping')
        return _Grouping(fields=fields)

    @classmethod
    def custom_object(cls, java_class_name, arg_list):
        """Tuples will be assigned to tasks by the given Java class."""
        java_object = storm_thrift.JavaObject(full_class_name=java_class_name,
                                              arg_list=[to_java_arg(arg)
                                                        for arg in arg_list])
        return _Grouping(custom_object=java_object)

    @classmethod
    def custom_serialized(cls, java_serialized):
        """Tuples will be assigned to tasks by the given Java serialized class.
        """
        if not isinstance(java_serialized, bytes):
            return TypeError('Argument to custom_serialized must be a '
                             'serialized Java class as bytes.  Given: {!r}'
                             .format(java_serialized))
        return _Grouping(custom_serialized=java_serialized)



"""
Python Storm Topology DSL
"""

import io

import six
import thriftpy

# Apache storm.thrift file as string to simplify loading it.
_THRIFT_STR = """
#!/usr/local/bin/thrift --gen java:beans,nocamel,hashcode

/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements. See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership. The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 *
 * Contains some contributions under the Thrift Software License.
 * Please see doc/old-thrift-license.txt in the Thrift distribution for
 * details.
 */

namespace java backtype.storm.generated

union JavaObjectArg {
  1: i32 int_arg;
  2: i64 long_arg;
  3: string string_arg;
  4: bool bool_arg;
  5: binary binary_arg;
  6: double double_arg;
}

struct JavaObject {
  1: required string full_class_name;
  2: required list<JavaObjectArg> args_list;
}

struct NullStruct {

}

struct GlobalStreamId {
  1: required string componentId;
  2: required string streamId;
  #Going to need to add an enum for the stream type (NORMAL or FAILURE)
}

union Grouping {
  1: list<string> fields; //empty list means global grouping
  2: NullStruct shuffle; // tuple is sent to random task
  3: NullStruct all; // tuple is sent to every task
  4: NullStruct none; // tuple is sent to a single task (storm's choice) -> allows storm to optimize the topology by bundling tasks into a single process
  5: NullStruct direct; // this bolt expects the source bolt to send tuples directly to it
  6: JavaObject custom_object;
  7: binary custom_serialized;
  8: NullStruct local_or_shuffle; // prefer sending to tasks in the same worker process, otherwise shuffle
}

struct StreamInfo {
  1: required list<string> output_fields;
  2: required bool direct;
}

struct ShellComponent {
  // should change this to 1: required list<string> execution_command;
  1: string execution_command;
  2: string script;
}

union ComponentObject {
  1: binary serialized_java;
  2: ShellComponent shell;
  3: JavaObject java_object;
}

struct ComponentCommon {
  1: required map<GlobalStreamId, Grouping> inputs;
  2: required map<string, StreamInfo> streams; //key is stream id
  3: optional i32 parallelism_hint; //how many threads across the cluster should be dedicated to this component

  // component specific configuration respects:
  // topology.debug: false
  // topology.max.task.parallelism: null // can replace isDistributed with this
  // topology.max.spout.pending: null
  // topology.kryo.register // this is the only additive one

  // component specific configuration
  4: optional string json_conf;
}

struct SpoutSpec {
  1: required ComponentObject spout_object;
  2: required ComponentCommon common;
  // can force a spout to be non-distributed by overriding the component configuration
  // and setting TOPOLOGY_MAX_TASK_PARALLELISM to 1
}

struct Bolt {
  1: required ComponentObject bolt_object;
  2: required ComponentCommon common;
}

// not implemented yet
// this will eventually be the basis for subscription implementation in storm
struct StateSpoutSpec {
  1: required ComponentObject state_spout_object;
  2: required ComponentCommon common;
}

struct StormTopology {
  //ids must be unique across maps
  // #workers to use is in conf
  1: required map<string, SpoutSpec> spouts;
  2: required map<string, Bolt> bolts;
  3: required map<string, StateSpoutSpec> state_spouts;
}

exception AlreadyAliveException {
  1: required string msg;
}

exception NotAliveException {
  1: required string msg;
}

exception AuthorizationException {
  1: required string msg;
}

exception InvalidTopologyException {
  1: required string msg;
}

struct TopologySummary {
  1: required string id;
  2: required string name;
  3: required i32 num_tasks;
  4: required i32 num_executors;
  5: required i32 num_workers;
  6: required i32 uptime_secs;
  7: required string status;
513: optional string sched_status;
514: optional string owner;
515: optional i32 replication_count;
}

struct SupervisorSummary {
  1: required string host;
  2: required i32 uptime_secs;
  3: required i32 num_workers;
  4: required i32 num_used_workers;
  5: required string supervisor_id;
  6: optional string version = "VERSION_NOT_PROVIDED";
  7: optional map<string, double> total_resources;
}

struct NimbusSummary {
  1: required string host;
  2: required i32 port;
  3: required i32 uptime_secs;
  4: required bool isLeader;
  5: required string version;
}

struct ClusterSummary {
  1: required list<SupervisorSummary> supervisors;
  3: required list<TopologySummary> topologies;
  4: required list<NimbusSummary> nimbuses;
}

struct ErrorInfo {
  1: required string error;
  2: required i32 error_time_secs;
  3: optional string host;
  4: optional i32 port;
}

struct BoltStats {
  1: required map<string, map<GlobalStreamId, i64>> acked;
  2: required map<string, map<GlobalStreamId, i64>> failed;
  3: required map<string, map<GlobalStreamId, double>> process_ms_avg;
  4: required map<string, map<GlobalStreamId, i64>> executed;
  5: required map<string, map<GlobalStreamId, double>> execute_ms_avg;
}

struct SpoutStats {
  1: required map<string, map<string, i64>> acked;
  2: required map<string, map<string, i64>> failed;
  3: required map<string, map<string, double>> complete_ms_avg;
}

union ExecutorSpecificStats {
  1: BoltStats bolt;
  2: SpoutStats spout;
}

// Stats are a map from the time window (all time or a number indicating number of seconds in the window)
//    to the stats. Usually stats are a stream id to a count or average.
struct ExecutorStats {
  1: required map<string, map<string, i64>> emitted;
  2: required map<string, map<string, i64>> transferred;
  3: required ExecutorSpecificStats specific;
  4: required double rate;
}

struct ExecutorInfo {
  1: required i32 task_start;
  2: required i32 task_end;
}

struct ExecutorSummary {
  1: required ExecutorInfo executor_info;
  2: required string component_id;
  3: required string host;
  4: required i32 port;
  5: required i32 uptime_secs;
  7: optional ExecutorStats stats;
}

struct DebugOptions {
  1: optional bool enable
  2: optional double samplingpct
}

struct TopologyInfo {
  1: required string id;
  2: required string name;
  3: required i32 uptime_secs;
  4: required list<ExecutorSummary> executors;
  5: required string status;
  6: required map<string, list<ErrorInfo>> errors;
  7: optional map<string, DebugOptions> component_debug;
513: optional string sched_status;
514: optional string owner;
515: optional i32 replication_count;
}

struct CommonAggregateStats {
1: optional i32 num_executors;
2: optional i32 num_tasks;
3: optional i64 emitted;
4: optional i64 transferred;
5: optional i64 acked;
6: optional i64 failed;
}

struct SpoutAggregateStats {
1: optional double complete_latency_ms;
}

struct BoltAggregateStats {
1: optional double execute_latency_ms;
2: optional double process_latency_ms;
3: optional i64    executed;
4: optional double capacity;
}

union SpecificAggregateStats {
1: BoltAggregateStats  bolt;
2: SpoutAggregateStats spout;
}

enum ComponentType {
  BOLT = 1,
  SPOUT = 2
}

struct ComponentAggregateStats {
1: optional ComponentType type;
2: optional CommonAggregateStats common_stats;
3: optional SpecificAggregateStats specific_stats;
4: optional ErrorInfo last_error;
}

struct TopologyStats {
1: optional map<string, i64> window_to_emitted;
2: optional map<string, i64> window_to_transferred;
3: optional map<string, double> window_to_complete_latencies_ms;
4: optional map<string, i64> window_to_acked;
5: optional map<string, i64> window_to_failed;
}

struct TopologyPageInfo {
 1: required string id;
 2: optional string name;
 3: optional i32 uptime_secs;
 4: optional string status;
 5: optional i32 num_tasks;
 6: optional i32 num_workers;
 7: optional i32 num_executors;
 8: optional string topology_conf;
 9: optional map<string,ComponentAggregateStats> id_to_spout_agg_stats;
10: optional map<string,ComponentAggregateStats> id_to_bolt_agg_stats;
11: optional string sched_status;
12: optional TopologyStats topology_stats;
13: optional string owner;
14: optional DebugOptions debug_options;
15: optional i32 replication_count;
}

struct ExecutorAggregateStats {
1: optional ExecutorSummary exec_summary;
2: optional ComponentAggregateStats stats;
}

struct ComponentPageInfo {
 1: required string component_id;
 2: required ComponentType component_type;
 3: optional string topology_id;
 4: optional string topology_name;
 5: optional i32 num_executors;
 6: optional i32 num_tasks;
 7: optional map<string,ComponentAggregateStats> window_to_stats;
 8: optional map<GlobalStreamId,ComponentAggregateStats> gsid_to_input_stats;
 9: optional map<string,ComponentAggregateStats> sid_to_output_stats;
10: optional list<ExecutorAggregateStats> exec_stats;
11: optional list<ErrorInfo> errors;
12: optional string eventlog_host;
13: optional i32 eventlog_port;
14: optional DebugOptions debug_options;
15: optional string topology_status;
}

struct KillOptions {
  1: optional i32 wait_secs;
}

struct RebalanceOptions {
  1: optional i32 wait_secs;
  2: optional i32 num_workers;
  3: optional map<string, i32> num_executors;
}

struct Credentials {
  1: required map<string,string> creds;
}

enum TopologyInitialStatus {
    ACTIVE = 1,
    INACTIVE = 2
}
struct SubmitOptions {
  1: required TopologyInitialStatus initial_status;
  2: optional Credentials creds;
}

struct SupervisorInfo {
    1: required i64 time_secs;
    2: required string hostname;
    3: optional string assignment_id;
    4: optional list<i64> used_ports;
    5: optional list<i64> meta;
    6: optional map<string, string> scheduler_meta;
    7: optional i64 uptime_secs;
    8: optional string version;
    9: optional map<string, double> resources_map;
}

struct NodeInfo {
    1: required string node;
    2: required set<i64> port;
}

struct Assignment {
    1: required string master_code_dir;
    2: optional map<string, string> node_host = {};
    3: optional map<list<i64>, NodeInfo> executor_node_port = {};
    4: optional map<list<i64>, i64> executor_start_time_secs = {};
}

enum TopologyStatus {
    ACTIVE = 1,
    INACTIVE = 2,
    REBALANCING = 3,
    KILLED = 4
}

union TopologyActionOptions {
    1: optional KillOptions kill_options;
    2: optional RebalanceOptions rebalance_options;
}

struct StormBase {
    1: required string name;
    2: required TopologyStatus status;
    3: required i32 num_workers;
    4: optional map<string, i32> component_executors;
    5: optional i32 launch_time_secs;
    6: optional string owner;
    7: optional TopologyActionOptions topology_action_options;
    8: optional TopologyStatus prev_status;//currently only used during rebalance action.
    9: optional map<string, DebugOptions> component_debug; // topology/component level debug option.
}

struct ClusterWorkerHeartbeat {
    1: required string storm_id;
    2: required map<ExecutorInfo,ExecutorStats> executor_stats;
    3: required i32 time_secs;
    4: required i32 uptime_secs;
}

struct ThriftSerializedObject {
  1: required string name;
  2: required binary bits;
}

struct LocalStateData {
   1: required map<string, ThriftSerializedObject> serialized_parts;
}

struct LocalAssignment {
  1: required string topology_id;
  2: required list<ExecutorInfo> executors;
}

struct LSSupervisorId {
   1: required string supervisor_id;
}

struct LSApprovedWorkers {
   1: required map<string, i32> approved_workers;
}

struct LSSupervisorAssignments {
   1: required map<i32, LocalAssignment> assignments;
}

struct LSWorkerHeartbeat {
   1: required i32 time_secs;
   2: required string topology_id;
   3: required list<ExecutorInfo> executors
   4: required i32 port;
}

enum NumErrorsChoice {
  ALL,
  NONE,
  ONE
}

struct GetInfoOptions {
  1: optional NumErrorsChoice num_err_choice;
}

enum LogLevelAction {
  UNCHANGED = 1,
  UPDATE    = 2,
  REMOVE    = 3
}

struct LogLevel {
  1: required LogLevelAction action;

  // during this thrift call, we'll move logger to target_log_level
  2: optional string target_log_level;

  // number of seconds that target_log_level should be kept
  // after this timeout, the loggers will be reset to reset_log_level
  // if timeout is 0, we will not reset
  3: optional i32 reset_log_level_timeout_secs;

  // number of seconds since unix epoch corresponding to
  // current time (when message gets to nimbus) + reset_log_level_timeout_se
  // NOTE: this field gets set in Nimbus
  4: optional i64 reset_log_level_timeout_epoch;

  // if reset timeout was set, then we would reset
  // to this level after timeout (or INFO by default)
  5: optional string reset_log_level;
}

struct LogConfig {
  // logger name -> log level map
  2: optional map<string, LogLevel> named_logger_level;
}

service Nimbus {
  void submitTopology(1: string name, 2: string uploadedJarLocation, 3: string jsonConf, 4: StormTopology topology) throws (1: AlreadyAliveException e, 2: InvalidTopologyException ite, 3: AuthorizationException aze);
  void submitTopologyWithOpts(1: string name, 2: string uploadedJarLocation, 3: string jsonConf, 4: StormTopology topology, 5: SubmitOptions options) throws (1: AlreadyAliveException e, 2: InvalidTopologyException ite, 3: AuthorizationException aze);
  void killTopology(1: string name) throws (1: NotAliveException e, 2: AuthorizationException aze);
  void killTopologyWithOpts(1: string name, 2: KillOptions options) throws (1: NotAliveException e, 2: AuthorizationException aze);
  void activate(1: string name) throws (1: NotAliveException e, 2: AuthorizationException aze);
  void deactivate(1: string name) throws (1: NotAliveException e, 2: AuthorizationException aze);
  void rebalance(1: string name, 2: RebalanceOptions options) throws (1: NotAliveException e, 2: InvalidTopologyException ite, 3: AuthorizationException aze);

  // dynamic log levels
  void setLogConfig(1: string name, 2: LogConfig config);
  LogConfig getLogConfig(1: string name);

  /**
  * Enable/disable logging the tuples generated in topology via an internal EventLogger bolt. The component name is optional
  * and if null or empty, the debug flag will apply to the entire topology.
  *
  * The 'samplingPercentage' will limit loggging to a percentage of generated tuples.
  **/
  void debug(1: string name, 2: string component, 3: bool enable, 4: double samplingPercentage) throws (1: NotAliveException e, 2: AuthorizationException aze);
  void uploadNewCredentials(1: string name, 2: Credentials creds) throws (1: NotAliveException e, 2: InvalidTopologyException ite, 3: AuthorizationException aze);

  // need to add functions for asking about status of storms, what nodes they're running on, looking at task logs

  string beginFileUpload() throws (1: AuthorizationException aze);
  void uploadChunk(1: string location, 2: binary chunk) throws (1: AuthorizationException aze);
  void finishFileUpload(1: string location) throws (1: AuthorizationException aze);

  string beginFileDownload(1: string file) throws (1: AuthorizationException aze);
  //can stop downloading chunks when receive 0-length byte array back
  binary downloadChunk(1: string id) throws (1: AuthorizationException aze);

  // returns json
  string getNimbusConf() throws (1: AuthorizationException aze);
  // stats functions
  ClusterSummary getClusterInfo() throws (1: AuthorizationException aze);
  TopologyInfo getTopologyInfo(1: string id) throws (1: NotAliveException e, 2: AuthorizationException aze);
  TopologyInfo getTopologyInfoWithOpts(1: string id, 2: GetInfoOptions options) throws (1: NotAliveException e, 2: AuthorizationException aze);
  TopologyPageInfo getTopologyPageInfo(1: string id, 2: string window, 3: bool is_include_sys) throws (1: NotAliveException e, 2: AuthorizationException aze);
  ComponentPageInfo getComponentPageInfo(1: string topology_id, 2: string component_id, 3: string window, 4: bool is_include_sys) throws (1: NotAliveException e, 2: AuthorizationException aze);
  //returns json
  string getTopologyConf(1: string id) throws (1: NotAliveException e, 2: AuthorizationException aze);
  /**
   * Returns the compiled topology that contains ackers and metrics consumsers. Compare {@link #getUserTopology(String id)}.
   */
  StormTopology getTopology(1: string id) throws (1: NotAliveException e, 2: AuthorizationException aze);
  /**
   * Returns the user specified topology as submitted originally. Compare {@link #getTopology(String id)}.
   */
  StormTopology getUserTopology(1: string id) throws (1: NotAliveException e, 2: AuthorizationException aze);
}

struct DRPCRequest {
  1: required string func_args;
  2: required string request_id;
}

exception DRPCExecutionException {
  1: required string msg;
}

service DistributedRPC {
  string execute(1: string functionName, 2: string funcArgs) throws (1: DRPCExecutionException e, 2: AuthorizationException aze);
}

service DistributedRPCInvocations {
  void result(1: string id, 2: string result) throws (1: AuthorizationException aze);
  DRPCRequest fetchRequest(1: string functionName) throws (1: AuthorizationException aze);
  void failRequest(1: string id) throws (1: AuthorizationException aze);
}
"""

io_class = io.StringIO if six.PY3 else io.BytesIO
storm_thrift = thriftpy.load_fp(io_class(_THRIFT_STR),
                                module_name='storm_thrift')
# Fix a lovely issue where GlobalStreamIds didn't have consistent hash values
storm_thrift.GlobalStreamId.__hash__ = lambda self: (hash(self.componentId) ^
                                                     hash(self.streamId))



"""
Package that adds streamparse-specific addition to pystorm classes
"""

from pystorm import Tuple

from .bolt import BatchingBolt, Bolt, JavaBolt, ShellBolt, TicklessBatchingBolt
from .component import Component, StormHandler
from .spout import JavaSpout, ShellSpout, Spout



"""
Functions for adding common CLI arguments to argparse sub-commands.
"""
import argparse
import copy

from ruamel import yaml
from six import integer_types, string_types

class _StoreDictAction(argparse.Action):
    """Action for storing key=val option strings as a single dict."""
    def __init__(self, option_strings, dest, nargs=None, const=None,
                 default=None, type=None, choices=None, required=False,
                 help=None, metavar=None):
        if nargs == 0:
            raise ValueError('nargs for store_dict actions must be > 0')
        if const is not None and nargs != '?':
            raise ValueError('nargs must be "?" to supply const')
        super(_StoreDictAction, self).__init__(option_strings=option_strings,
                                               dest=dest,
                                               nargs=nargs,
                                               const=const,
                                               default=default,
                                               type=type,
                                               choices=choices,
                                               required=required,
                                               help=help,
                                               metavar=metavar)

    def __call__(self, parser, namespace, values, option_string=None):
        if getattr(namespace, self.dest, None) is None:
            setattr(namespace, self.dest, {})
        # Only doing a copy here because that's what _AppendAction does
        items = copy.copy(getattr(namespace, self.dest))
        key, val = values.split("=", 1)
        items[key] = yaml.load(val)
        setattr(namespace, self.dest, items)


def option_alias(option):
    """Returns a function that will create option=val for _StoreDictAction."""
    def _create_key_val_str(val):
        return '{}={}'.format(option, val)
    return _create_key_val_str


def add_ackers(parser):
    """ Add --ackers option to parser """
    parser.add_argument('-a', '--ackers',
                        help='Set number of acker executors for your topology. '
                             'Defaults to the number of worker nodes in your '
                             'Storm environment.',
                        type=option_alias('topology.acker.executors'),
                        action=_StoreDictAction,
                        dest='options')


def add_debug(parser):
    """ Add --debug option to parser """
    parser.add_argument('-d', '--debug',
                        help='Set topology.debug and produce debugging output.',
                        type=option_alias('topology.debug'),
                        action=_StoreDictAction,
                        dest='options',
                        const='true',
                        nargs='?')


def add_environment(parser):
    """ Add --environment option to parser """
    parser.add_argument('-e', '--environment',
                        help='The environment to use for the command.  '
                             'Corresponds to an environment in your '
                             '"envs" dictionary in config.json.  If you '
                             'only have one environment specified, '
                             'streamparse will automatically use this.')


def add_name(parser):
    """ Add --name option to parser """
    parser.add_argument('-n', '--name',
                        help='The name of the topology to act on.  If you have '
                             'only one topology defined in your "topologies" '
                             'directory, streamparse will use it '
                             'automatically.')


def add_options(parser):
    """ Add --option options to parser """
    parser.add_argument('-o', '--option',
                        dest='options',
                        action=_StoreDictAction,
                        help='Topology option to pass on to Storm. For example,'
                             ' "-o topology.debug=true" is equivalent to '
                             '"--debug".  May be repeated multiple for multiple'
                             ' options.')


def add_override_name(parser):
    """ Add --override_name option to parser """
    parser.add_argument('-N', '--override_name',
                        help='For operations such as creating virtualenvs and '
                             'killing/submitting topologies, use this value '
                             'instead of NAME.  This is useful if you want to '
                             'submit the same topology twice without having to '
                             'duplicate the topology file.')


def add_pattern(parser):
    """ Add --pattern option to parser """
    parser.add_argument('--pattern',
                        help='Pattern of log files to operate on.')


def add_simple_jar(parser):
    """ Add --simple_jar option to parser. """
    parser.add_argument("-s", "--simple_jar",
                        action='store_true',
                        help='Instead of creating an Uber-JAR for the '
                             'topology, which contains all of its JVM '
                             'dependencies, create a simple JAR with just the '
                             'code for the project.  This is useful when your '
                             'project is pure Python and has no JVM '
                             'dependencies.')


def add_wait(parser):
    """ Add --wait option to parser """
    parser.add_argument('--wait',
                        type=int,
                        default=5,
                        help='Seconds to wait before killing topology. '
                             '(default: %(default)s)')


def add_workers(parser):
    """ Add --workers option to parser """
    parser.add_argument('-w', '--workers',
                        help='Set number of Storm workers for your topology. '
                             'Defaults to the number of worker nodes in your '
                             'Storm environment.',
                        type=option_alias('topology.workers'),
                        action=_StoreDictAction,
                        dest='options')


def resolve_options(cli_options, env_config, topology_class, topology_name):
    """Resolve potentially conflicting Storm options from three sources:

    CLI options > Topology options > config.json options
    """
    storm_options = {}

    # Start with environment options
    storm_options.update(env_config.get('options', {}))

    # Set topology.python.path
    if env_config.get('use_virtualenv', True):
        python_path = '/'.join([env_config["virtualenv_root"],
                                topology_name, "bin", "python"])
        # This setting is for information purposes only, and is not actually
        # read by any pystorm or streamparse code.
        storm_options['topology.python.path'] = python_path

    # Set logging options based on environment config
    log_config = env_config.get("log", {})
    log_path = log_config.get("path") or env_config.get("log_path")
    log_file = log_config.get("file") or env_config.get("log_file")
    if log_path:
        storm_options['pystorm.log.path'] = log_path
    if log_file:
        storm_options['pystorm.log.file'] = log_file
    if isinstance(log_config.get("max_bytes"), integer_types):
        storm_options['pystorm.log.max_bytes'] = log_config["max_bytes"]
    if isinstance(log_config.get("backup_count"), integer_types):
        storm_options['pystorm.log.backup_count'] = log_config["backup_count"]
    if isinstance(log_config.get("level"), string_types):
        storm_options['pystorm.log.level'] = log_config["level"].lower()

    # Override options with topology options
    storm_options.update(topology_class.config)

    # Override options with CLI options
    storm_options.update(cli_options or {})

    # Set log level to debug if topology.debug is set
    if storm_options.get('topology.debug', False):
        storm_options['pystorm.log.level'] = 'debug'

    # If ackers and executors still aren't set, use number of worker nodes
    num_storm_workers = len(env_config["workers"])
    if storm_options.get('topology.acker.executors') is None:
        storm_options['topology.acker.executors'] = num_storm_workers
    if storm_options.get('topology.workers') is None:
        storm_options['topology.workers'] = num_storm_workers

    return storm_options
